from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABCMeta, abstractmethod
from six import add_metaclass

from itertools import combinations

from boltons.cacheutils import LRU
import numpy as np
from scipy.ndimage import zoom
from scipy.special import logsumexp
from tqdm import tqdm

from .generics import progressinfo
from .saliency_map_models import (SaliencyMapModel, handle_stimulus,
                                  SubjectDependentSaliencyMapModel,
                                  ExpSaliencyMapModel,
                                  DisjointUnionMixin
                                  )
from .datasets import FixationTrains, get_image_hash, as_stimulus
from .sampling_models import SamplingModelMixin
from .utils import Cache, average_values, deprecated_class, remove_trailing_nans
from .tf_utils import tf_logsumexp


def sample_from_logprobabilities(log_probabilities, size=1, rst=None):
    """ Sample from log probabilities (robust to many bins and small probabilities).

        +-np.inf and np.nan will be interpreted as zero probability
    """
    if rst is None:
        rst = np.random
    log_probabilities = np.asarray(log_probabilities)

    valid_indices = np.nonzero(np.isfinite(log_probabilities))[0]
    valid_log_probabilities = log_probabilities[valid_indices]

    ndxs = valid_log_probabilities.argsort()
    sorted_log_probabilities = valid_log_probabilities[ndxs]
    cumsums = np.logaddexp.accumulate(sorted_log_probabilities)
    cumsums -= cumsums[-1]

    tmps = -rst.exponential(size=size)
    js = np.searchsorted(cumsums, tmps)
    valid_values = ndxs[js]
    values = valid_indices[valid_values]

    return values


def sample_from_logdensity(log_density, count=None, rst=None):
    if count is None:
        real_count = 1
    else:
        real_count = count

    height, width = log_density.shape
    flat_log_density = log_density.flatten(order='C')
    samples = sample_from_logprobabilities(flat_log_density, size=real_count, rst=rst)
    sample_xs = samples % width
    sample_ys = samples // width

    if count is None:
        return sample_xs[0], sample_ys[0]
    else:
        return np.asarray(sample_xs), np.asarray(sample_ys)


def sample_from_image(densities, count=None, rst=None):
    if rst is None:
        rst = np.random
    height, width = densities.shape
    sorted_densities = densities.flatten(order='C')
    cumsums = np.cumsum(sorted_densities)
    if count is None:
        real_count = 1
    else:
        real_count = count
    sample_xs = []
    sample_ys = []
    tmps = rst.rand(real_count)
    js = np.searchsorted(cumsums, tmps)
    for j in js:
        sample_xs.append(j % width)
        sample_ys.append(j // width)
    sample_xs = np.asarray(sample_xs)
    sample_ys = np.asarray(sample_ys)
    if count is None:
        return sample_xs[0], sample_ys[0]
    else:
        return np.asarray(sample_xs), np.asarray(sample_ys)


@add_metaclass(ABCMeta)
class ScanpathModel(SamplingModelMixin, object):
    """
    General probabilistic saliency model.

    Inheriting classes have to implement `conditional_log_density`
    """

    @abstractmethod
    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None):
        raise NotImplementedError()

    def conditional_log_density_for_fixation(self, stimuli, fixations, fixation_index, out=None):
        return self.conditional_log_density(
            stimuli.stimulus_objects[fixations.n[fixation_index]],
            x_hist=remove_trailing_nans(fixations.x_hist[fixation_index]),
            y_hist=remove_trailing_nans(fixations.y_hist[fixation_index]),
            t_hist=remove_trailing_nans(fixations.t_hist[fixation_index]),
            attributes={key: getattr(fixations, key)[fixation_index] for key in fixations.__attributes__},
            out=out
        )

    def log_likelihoods(self, stimuli, fixations, verbose=False):
        log_likelihoods = np.empty(len(fixations.x))
        for i in tqdm(range(len(fixations.x)), disable=not verbose):
            conditional_log_density = self.conditional_log_density_for_fixation(stimuli, fixations, i)
            log_likelihoods[i] = conditional_log_density[fixations.y_int[i], fixations.x_int[i]]

        return log_likelihoods

    def log_likelihood(self, stimuli, fixations, verbose=False, average='fixation'):
        log_likelihoods = self.log_likelihoods(stimuli, fixations, verbose=verbose)


        return average_values(self.log_likelihoods(stimuli, fixations, verbose=verbose), fixations, average=average)

    def information_gains(self, stimuli, fixations, baseline_model=None, verbose=False, average='fixation'):
        if baseline_model is None:
            baseline_model = UniformModel()

        own_log_likelihoods = self.log_likelihoods(stimuli, fixations, verbose=verbose)
        baseline_log_likelihoods = baseline_model.log_likelihoods(stimuli, fixations, verbose=verbose)
        return (own_log_likelihoods - baseline_log_likelihoods) / np.log(2)

    def information_gain(self, stimuli, fixations, baseline_model=None, verbose=False, average='fixation'):
        return average_values(self.information_gains(stimuli, fixations, baseline_model, verbose=verbose), fixations, average=average)

    def _expand_sample_arguments(self, stimuli, train_counts, lengths=None, stimulus_indices=None):
        if isinstance(train_counts, int):
            train_counts = [train_counts]*len(stimuli)

        if stimulus_indices is None:
            if len(train_counts) > len(stimuli):
                raise ValueError('Number of train counts higher than count of stimuli!')
            stimulus_indices = range(len(train_counts))

        if isinstance(stimulus_indices, int):
            stimulus_indices = [stimulus_indices]

        if len(stimulus_indices) != len(train_counts):
            raise ValueError('Number of train counts must match number of stimulus_indices!')

        if isinstance(lengths, int):
            lengths = [lengths for i in range(len(stimuli))]

        if len(train_counts) != len(lengths):
            raise ValueError('Number of train counts and number of lengths do not match!')

        new_lengths = []
        for k, l in enumerate(lengths):
            if isinstance(l, int):
                ll = [l for i in range(train_counts[k])]
            else:
                ll = l
            new_lengths.append(ll)
        lengths = new_lengths

        for k, (c, ls) in enumerate(zip(train_counts, lengths)):
            if c != len(ls):
                raise ValueError('{}th train count ({}) does not match given number of lengths ({})!'.format(k, c, len(ls)))

        return stimuli, train_counts, lengths, stimulus_indices

    def sample(self, stimuli, train_counts, lengths=1, stimulus_indices=None):
        """
        Sample fixations for given stimuli


        Examples
        --------

        >>> model.sample(stimuli, 10)  # sample 10 fixations per image
        >>> # sample 5 fixations from the first image, 10 from the second and
        >>> # 30 fixations from the third image
        >>> model.sample(stimuli, [5, 10, 30])
        >>> # Sample 10 fixation trains per image, each consiting of 5 fixations
        >>> model.sample(stimuli, 10, lengths=5)
        >>> # Sample 10 fixation trains per image, consisting of 2 fixations
        >>> # for the first image, 4 fixations for the second, ...
        >>> model.sample(stimuli, 10, lengths=[2, 4, 3])
        >>> # Sample
        >>> model.sample(stimuli, 3, lengths=[[1,2,3], [1,2,3]])

        >>> # Sample 3 fixations from the 10th stimulus
        >>> model.sample(stimuli, 3, stimulus_indices = 10)

        >>> # Sample 3 fixations from the 20th and the 42th stimuli each
        >>> model.sample(stimuli, 3, stimulus_indices = [20, 42])
        """

        stimuli, train_counts, lengths, stimulus_indices = self._expand_sample_arguments(stimuli,
                                                                                         train_counts,
                                                                                         lengths,
                                                                                         stimulus_indices)

        xs = []
        ys = []
        ts = []
        ns = []
        subjects = []
        for stimulus_index, ls in progressinfo(zip(stimulus_indices, lengths)):
            stimulus = stimuli[stimulus_index]
            for l in ls:
                this_xs, this_ys, this_ts = self._sample_fixation_train(stimulus, l)
                xs.append(this_xs)
                ys.append(this_ys)
                ts.append(this_ts)
                ns.append(stimulus_index)
                subjects.append(0)
        return FixationTrains.from_fixation_trains(xs, ys, ts, ns, subjects)

    def _sample_fixation_train(self, stimulus, length):
        """Sample one fixation train of given length from stimulus"""
        return self.sample_scanpath(stimulus, [], [], [], length)

    def sample_fixation(self, stimulus, x_hist, y_hist, t_hist, attributes=None, verbose=False, rst=None):
        log_densities = self.conditional_log_density(stimulus, x_hist, y_hist, t_hist, attributes=attributes)
        x, y = sample_from_image(np.exp(log_densities))
        return x, y, len(t_hist)


class Model(ScanpathModel):
    """
    Time independend probabilistic saliency model.

    Inheriting classes have to implement `_log_density`.
    """
    def __init__(self, cache_location=None, caching=True, memory_cache_size=None):
        super(Model, self).__init__()
        self._cache = Cache(cache_location, memory_cache_size=memory_cache_size)
        self.caching = caching
        #self._log_density_cache = Cache(cache_location)
        # This make the property `cache_location` work.
        #self._saliency_map_cache = self._log_density_cache

    @property
    def cache_location(self):
        return self._cache.cache_location

    @cache_location.setter
    def cache_location(self, value):
        self._cache.cache_location = value

    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None):
        return self.log_density(stimulus)

    def log_density(self, stimulus):
        """
        Get log_density for given stimulus.

        To overwrite this function, overwrite `_log_density` as otherwise
        the caching mechanism is disabled.
        """
        stimulus = handle_stimulus(stimulus)
        if not self.caching:
            return self._log_density(stimulus.stimulus_data)
        stimulus_id = stimulus.stimulus_id
        if not stimulus_id in self._cache:
            self._cache[stimulus_id] = self._log_density(stimulus.stimulus_data)
        return self._cache[stimulus_id]

    @abstractmethod
    def _log_density(self, stimulus):
        """
        Overwrite this to implement you own SaliencyMapModel.

        Parameters
        ----------

        @type  stimulus: ndarray
        @param stimulus: stimulus for which the saliency map should be computed.
        """
        raise NotImplementedError()

    def log_likelihoods(self, stimuli, fixations, verbose=False):
        log_likelihoods = np.empty(len(fixations.x))
        for n in tqdm(range(len(stimuli)), disable=not verbose):
            inds = fixations.n == n
            if not inds.sum():
                continue
            log_density = self.log_density(stimuli.stimulus_objects[n])
            this_log_likelihoods = log_density[fixations.y_int[inds], fixations.x_int[inds]]
            log_likelihoods[inds] = this_log_likelihoods

        return log_likelihoods

    def _sample_fixation_train(self, stimulus, length):
        """Sample one fixation train of given length from stimulus"""
        # We could reuse the implementation from `ScanpathModel`
        # but this implementation is much faster for long trains.
        log_densities = self.log_density(stimulus)
        xs, ys = sample_from_image(np.exp(log_densities), count=length)
        ts = np.arange(len(xs))
        return xs, ys, ts

    def pixel_space_information_gain(self, baseline, gold_standard, stimulus, eps=1e-20):
        log_p_gold = gold_standard.log_density(stimulus)
        log_p_baseline = baseline.log_density(stimulus)
        log_p_model = self.log_density(stimulus)
        p_gold = np.exp(log_p_gold)
        p_gold[p_gold == 0] = p_gold[p_gold > 0].min()
        ig = (p_gold)*(np.logaddexp(log_p_model, np.log(eps))-np.logaddexp(log_p_baseline, np.log(eps)))
        return ig

    def kl_divergences(self, stimuli, gold_standard, log_regularization=0, quotient_regularization=0, verbose=False):
        """Calculate KL Divergence between model and gold standard for each stimulus.

        This metric works only for probabilistic models.
        For the existing saliency metrics known as KL Divergence, see
        `image_based_kl_divergence` and `fixation_based_kl_divergence`.

        log_regularization and quotient_regularization are regularization constants that are used as in
        kldiv(p1, p2) = sum(p1*log(log_regularization + p1 / (p2 + quotient_regularization))).
        """
        assert isinstance(self, Model)
        assert isinstance(gold_standard, Model)

        def _kl_div(logp1, logp2, log_regularization=0, quotient_regularization=0):
            if log_regularization or quotient_regularization:
                return (np.exp(logp2)*np.log(log_regularization+np.exp(logp2)/(np.exp(logp1)+quotient_regularization))).sum()
            else:
                return (np.exp(logp2)*(logp2 - logp1)).sum()

        kl_divs = []
        for s in tqdm(stimuli, disable=not verbose):
            logp_model = self.log_density(s)
            logp_gold = gold_standard.log_density(s)
            kl_divs.append(_kl_div(logp_model, logp_gold, log_regularization=log_regularization, quotient_regularization=quotient_regularization))

        return kl_divs

    def set_params(self, **kwargs):
        """
	        Set model parameters, if the model has parameters

	        This method has to reset caches etc., if the depend on the parameters
        """
        if kwargs:
            raise ValueError('Unkown parameters!', kwargs)


class CachedModel(Model):
    """Density model which uses only precached densities
    """
    def __init__(self, cache_location, **kwargs):
        if cache_location is None:
            raise ValueError("CachedModel needs a cache location!")
        super(CachedModel, self).__init__(cache_location=cache_location, **kwargs)

    def _log_density(self, stimulus):
        raise NotImplementedError()


class UniformModel(Model):
    """Saliency model assuming uniform fixation distribution over space
    """
    def _log_density(self, stimulus):
        return np.zeros((stimulus.shape[0], stimulus.shape[1])) - np.log(stimulus.shape[0]) - np.log(stimulus.shape[1])

    def log_likelihoods(self, stimuli, fixations, verbose=False):
        lls = []
        for n in fixations.n:
            lls.append(-np.log(stimuli.shapes[n][0]) - np.log(stimuli.shapes[n][1]))
        return np.array(lls)


class MixtureModel(Model):
    """ A saliency model being a weighted mixture of a number of other models
    """
    def __init__(self, models, weights=None, **kwargs):
        """Create a mixture model from a list of models and a list of weights

           :param models: list of `Model` instances
           :param weights: list of weights for the different models. Do not have
                           to sum up to one, they will be normalized.
                           If `None`, will be set to a uniform mixture.
        """
        super(MixtureModel, self).__init__(**kwargs)
        self.models = models
        if weights is None:
            weights = np.ones(len(self.models))
        weights = np.asarray(weights, dtype=float)
        weights /= weights.sum()
        if not len(weights) == len(models):
            raise ValueError('models and weights must have same length!')
        self.weights = weights

    def _log_density(self, stimulus):
        log_densities = []
        for i, model in enumerate(self.models):
            log_density = model.log_density(stimulus).copy()
            log_density += np.log(self.weights[i])
            log_densities.append(log_density)

        log_density = logsumexp(log_densities, axis=0)
        np.testing.assert_allclose(np.exp(log_density).sum(), 1.0, rtol=1e-7)
        if not log_density.shape == (stimulus.shape[0], stimulus.shape[1]):
            raise ValueError('wrong density shape in mixture model! stimulus shape: ({}, {}), density shape: {}'.format(stimulus.shape[0], stimulus.shape[1], log_density.shape))
        return log_density


class ResizingModel(Model):
    def __init__(self, parent_model, verbose=True, **kwargs):
        if 'caching' not in kwargs:
            kwargs['caching'] = False
        self.verbose = verbose
        super(ResizingModel, self).__init__(**kwargs)
        self.parent_model = parent_model

    def _log_density(self, stimulus):
        smap = self.parent_model.log_density(stimulus)

        target_shape = (stimulus.shape[0],
                        stimulus.shape[1])

        if smap.shape != target_shape:
            if self.verbose:
                print("Resizing saliency map", smap.shape, target_shape)
            x_factor = target_shape[1] / smap.shape[1]
            y_factor = target_shape[0] / smap.shape[0]

            smap = zoom(smap, [y_factor, x_factor], order=1, mode='nearest')

            smap -= logsumexp(smap)

            assert smap.shape == target_shape

        return smap


class ResizingScanpathModel(ScanpathModel):
    def __init__(self, parent_model, verbose=True, **kwargs):
        self.verbose = verbose
        super(ResizingScanpathModel, self).__init__(**kwargs)
        self.parent_model = parent_model

    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None):
        smap = self.parent_model.conditional_log_density(stimulus, x_hist, y_hist, t_hist, attributes=attributes, out=out)

        target_shape = (stimulus.shape[0],
                        stimulus.shape[1])

        if smap.shape != target_shape:
            if self.verbose:
                print("Resizing saliency map", smap.shape, target_shape)
            x_factor = target_shape[1] / smap.shape[1]
            y_factor = target_shape[0] / smap.shape[0]

            smap = zoom(smap, [y_factor, x_factor], order=1, mode='nearest')

            smap -= logsumexp(smap)

            assert smap.shape == target_shape

        return smap


class DisjointUnionModel(DisjointUnionMixin, ScanpathModel):
    def conditional_log_density(self, stimulus, *args, **kwargs):
        raise

    def log_likelihoods(self, stimuli, fixations, **kwargs):
        return self.eval_metric('log_likelihoods', stimuli, fixations, **kwargs)


class SubjectDependentModel(DisjointUnionModel):
    def __init__(self, subject_models, **kwargs):
        super(SubjectDependentModel, self).__init__(**kwargs)
        self.subject_models = subject_models

    def _split_fixations(self, stimuli, fixations):
        for s in self.subject_models:
            yield fixations.subjects == s, self.subject_models[s]

    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None, **kwargs):
        if 'subjects' not in attributes:
            raise ValueError("SubjectDependentModel can't compute conditional log densities without subject indication!")
        return self.subject_models[attributes['subjects']].conditional_log_density(
            stimulus, x_hist, y_hist, t_hist, attributes=attributes, **kwargs)

    def get_saliency_map_model_for_sAUC(self, baseline_model):
        return SubjectDependentSaliencyMapModel({
            s: ShuffledAUCSaliencyMapModel(self.subject_models[s], baseline_model)
            for s in self.subject_models})

    def get_saliency_map_model_for_NSS(self):
        return SubjectDependentSaliencyMapModel({
            s: ExpSaliencyMapModel(self.subject_models[s])
            for s in self.subject_models})


class StimulusDependentModel(Model):
    def __init__(self, stimuli_models, check_stimuli=True, **kwargs):
        super(StimulusDependentModel, self).__init__(**kwargs)
        self.stimuli_models = stimuli_models
        if check_stimuli:
            self.check_stimuli()

    def check_stimuli(self):
        for s1, s2 in tqdm(list(combinations(self.stimuli_models, 2))):
            if not set(s1.stimulus_ids).isdisjoint(s2.stimulus_ids):
                raise ValueError('Stimuli not disjoint')

    def _log_density(self, stimulus):
        stimulus_hash = get_image_hash(stimulus)
        for stimuli, model in self.stimuli_models.items():
            if stimulus_hash in stimuli.stimulus_ids:
                return model.log_density(stimulus)
        else:
            raise ValueError('stimulus not provided by these models')


class StimulusDependentScanpathModel(ScanpathModel):
    def __init__(self, stimuli_models, check_stimuli=True, **kwargs):
        super(StimulusDependentScanpathModel, self).__init__(**kwargs)
        self.stimuli_models = stimuli_models
        if check_stimuli:
            self.check_stimuli()

    def check_stimuli(self):
        for s1, s2 in tqdm(list(combinations(self.stimuli_models, 2))):
            if not set(s1.stimulus_ids).isdisjoint(s2.stimulus_ids):
                raise ValueError('Stimuli not disjoint')

    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None):
        stimulus_hash = get_image_hash(as_stimulus(stimulus).stimulus_data)
        for stimuli, model in self.stimuli_models.items():
            if stimulus_hash in stimuli.stimulus_ids:
                return model.conditional_log_density(stimulus, x_hist, y_hist, t_hist, attributes=attributes, out=out)
        else:
            raise ValueError('stimulus not provided by these models')


class ShuffledAUCSaliencyMapModel(SaliencyMapModel):
    def __init__(self, probabilistic_model, baseline_model):
        super(ShuffledAUCSaliencyMapModel, self).__init__(caching=False)
        self.probabilistic_model = probabilistic_model
        self.baseline_model = baseline_model

    def _saliency_map(self, stimulus):
        return self.probabilistic_model.log_density(stimulus) - self.baseline_model.log_density(stimulus)


class ShuffledBaselineModel(Model):
    """Predicts a mixture of all predictions for other images.
    
    This model will usually be used as baseline model for computing sAUC saliency maps.
    """
    def __init__(self, parent_model, stimuli, resized_predictions_cache_size=5000,
                 compute_size=(500, 500),
                 **kwargs):
        super(ShuffledBaselineModel, self).__init__(**kwargs)
        self.parent_model = parent_model
        self.stimuli = stimuli
        self.compute_size = compute_size
        self.resized_predictions_cache = LRU(
            max_size=resized_predictions_cache_size,
            on_miss=self._cache_miss
        )

    def _resize_prediction(self, prediction, target_shape):
        if prediction.shape != target_shape:
            x_factor = target_shape[1] / prediction.shape[1]
            y_factor = target_shape[0] / prediction.shape[0]

            prediction = zoom(prediction, [y_factor, x_factor], order=1, mode='nearest')

            prediction -= logsumexp(prediction)

            assert prediction.shape == target_shape

        return prediction

    def _cache_miss(self, key):
        stimulus = self.stimuli[key]
        return self._resize_prediction(self.parent_model.log_density(stimulus), self.compute_size)

    def _log_density(self, stimulus):
        stimulus_id = get_image_hash(stimulus)

        predictions = []
        prediction = None
        count = 0

        target_shape = (stimulus.shape[0], stimulus.shape[1])

        for k, other_stimulus in enumerate((self.stimuli)):
            if other_stimulus.stimulus_id == stimulus_id:
                continue
            other_prediction = self.resized_predictions_cache[k]
            predictions.append(other_prediction)

        predictions = np.array(predictions) - np.log(len(predictions))
        prediction = tf_logsumexp(predictions, axis=0)

        prediction = self._resize_prediction(prediction, target_shape)

        return prediction


GeneralModel = deprecated_class(deprecated_in='0.2.16', removed_in='1.0.0', details="Use ScanpathModel instead")(ScanpathModel)
StimulusDependentGeneralModel = deprecated_class(deprecated_in='0.2.16', removed_in='1.0.0', details="Use StimulusDependentScanpathModel instead")(StimulusDependentScanpathModel)



from __future__ import division, print_function

import numpy as np

from boltons.iterutils import chunked

from .datasets import create_subset


def train_split(stimuli, fixations, crossval_folds, fold_no, val_folds=1, test_folds=1, random=True):
    return crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='train')


def validation_split(stimuli, fixations, crossval_folds, fold_no, val_folds=1, test_folds=1, random=True):
    return crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='val')


def test_split(stimuli, fixations, crossval_folds, fold_no, val_folds=1, test_folds=1, random=True):
    return crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='test')


def crossval_splits(stimuli, fixations, crossval_folds, fold_no, val_folds=1, test_folds=1, random=True):
    return (
        crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='train'),
        crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='val'),
        crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=val_folds, test_folds=test_folds, random=random, split='test'),
    )


def crossval_split(stimuli, fixations, crossval_folds, fold_no, val_folds=1, test_folds=1, random=True, split='train'):
    train_folds, val_folds, test_folds = get_crossval_folds(crossval_folds, fold_no, test_folds=test_folds, val_folds=val_folds)

    if split == 'train':
        folds = train_folds
    elif split == 'val':
        folds = val_folds
    elif split == 'test':
        folds = test_folds
    else:
        raise ValueError(split)

    return _get_crossval_split(stimuli, fixations, crossval_folds, included_splits=folds, random=random)


def _get_crossval_split(stimuli, fixations, split_count, included_splits, random=True):
    inds = list(range(len(stimuli)))
    if random:
        print("Using random shuffles for crossvalidation")
        rst = np.random.RandomState(seed=42)
        rst.shuffle(inds)
        inds = list(inds)
    size = int(np.ceil(len(inds) / split_count))
    chunks = chunked(inds, size=size)

    inds = []
    for split_nr in included_splits:
        inds.extend(chunks[split_nr])

    stimuli, fixations = create_subset(stimuli, fixations, inds)
    return stimuli, fixations


def create_train_folds(crossval_folds, val_folds, test_folds):
    all_folds = list(range(crossval_folds))
    if isinstance(val_folds, int):
        val_folds = [val_folds]
    if isinstance(test_folds, int):
        test_folds = [test_folds]

    train_folds = [f for f in all_folds if not (f in val_folds or f in test_folds)]

    return train_folds, val_folds, test_folds


def get_crossval_folds(crossval_folds, crossval_no, test_folds=1, val_folds=1):
    assert test_folds <= 1
    if test_folds:
        _test_folds = [crossval_no]
        _val_folds = [(crossval_no - i - 1) % crossval_folds for i in range(val_folds)]

    else:
        assert val_folds == 1

        _test_folds = [crossval_no]
        _val_folds = [crossval_no]

    _train_folds, _val_folds, _test_folds = create_train_folds(crossval_folds, _val_folds, _test_folds)

    return _train_folds, _val_folds, _test_folds


def iterate_crossvalidation(stimuli, fixations, crossval_folds, val_folds=1, test_folds=1, random=True):
    """iterate over crossvalidation folds. Each fold will yield
          train_stimuli, train_fixations, val_, test_stimuli, test_fixations
    """
    kwargs = {
        'crossval_folds': crossval_folds,
        'val_folds': val_folds,
        'test_folds': test_folds,
        'random': random
    }

    for fold_no in range(crossval_folds):
        train_stimuli, train_fixations = train_split(
            stimuli, fixations,
            fold_no=fold_no,
            **kwargs)
        val_stimuli, val_fixations = validation_split(
            stimuli, fixations,
            fold_no=fold_no,
            **kwargs)
        test_stimuli, test_fixations = test_split(
            stimuli, fixations,
            fold_no=fold_no,
            **kwargs)

        yield train_stimuli, train_fixations, val_stimuli, val_fixations, test_stimuli, test_fixations


def parse_list_of_intervals(description):
    """parses a string as "1.0:3.0,5.0:5.6,7" into [(1.0, 3.0), (5.0, 5.6), (7,)]
    """
    intervals = description.split(',')
    return [parse_interval(interval) for interval in intervals]


def parse_interval(interval):
    parts = interval.split(':')
    if len(parts) not in [1, 2]:
        raise ValueError("Invalid interval", interval)
    return tuple([float(part.strip()) for part in parts])


def filter_fixations_by_number(fixations, intervals):
    intervals = _check_intervals(intervals, type=int)
    inds = np.zeros_like(fixations.x, dtype=bool)

    for n1, n2 in intervals:
        _inds = np.logical_and(fixations.lengths >= n1, fixations.lengths < n2)
        inds = np.logical_or(inds, _inds)

    return fixations[inds]


def filter_stimuli_by_number(stimuli, fixations, intervals):
    intervals = _check_intervals(intervals, type=int)
    mask = np.zeros(len(stimuli), dtype=bool)

    for n1, n2 in intervals:
        mask[n1:n2] = True

    indices = list(np.nonzero(mask)[0])

    return create_subset(stimuli, fixations, indices)


def _check_intervals(intervals, type=float):
    if isinstance(intervals, (float, int)):
        intervals = [intervals]

    new_intervals = []
    for interval in intervals:
        new_intervals.append(_check_interval(interval, type=type))
    return new_intervals


def _check_interval(interval, type=float):
    if isinstance(interval, (float, int)):
        interval = [interval]

    if len(interval) == 1:
        if type != int:
            raise ValueError("single-value intervals only allowed for integer data!")
        interval = [interval[0], interval[0] + 1]

    if len(interval) != 2:
        raise ValueError("Intervals need two values", interval)
    new_interval = []
    for value in interval:
        if type(value) != value:
            raise ValueError("Invalid value for this type", value, type)
        new_interval.append(type(value))

    return tuple(new_interval)



#%%cython
# Circumvents a bug(?) in cython:
# http://stackoverflow.com/a/13976504
STUFF = "Hi"


import numpy as np
cimport numpy as np
cimport cython


#Do not check for index errors
@cython.boundscheck(False)
#Do not enable negativ indices
@cython.wraparound(False)
#Use native c division
@cython.cdivision(True)
def real_ROC(image, fixation_data, int judd=0):
    fixations_orig = np.zeros_like(image)
    fixations_orig[fixation_data] = 1.0
    image_1d = image.flatten()
    cdef np.ndarray[double, ndim=1] fixations = fixations_orig.flatten()
    inds = image_1d.argsort()
    #image_1d = image_1d[inds]
    fixations = fixations[inds]
    cdef np.ndarray[double, ndim=1] image_sorted = image_1d[inds]
    cdef np.ndarray[double, ndim=1] fixation_values_sorted = image[fixation_data]
    fixation_values_sorted.sort()
    cdef int i
    cdef int N = image_1d.shape[0]
    cdef int fix_count = fixations.sum()
    cdef int false_count = N-fix_count
    cdef int correct_count = 0
    cdef int false_positive_count = 0
    cdef int length
    if judd:
        length = fix_count+2
        assert len(fixation_values_sorted) == fix_count
    else:
        length = N+1
    cdef np.ndarray[double, ndim=1] precs = np.zeros(length)
    cdef np.ndarray[double, ndim=1] false_positives = np.zeros(length)
    for i in range(N):
        #print fixations[N-i-1],
        #print image_1d[N-i-1]
        # Every pixel is a nonfixation
        false_positive_count += 1
        if fixations[N-i-1]:
            correct_count += 1
            if judd:
                if i == N - 1 or fixation_values_sorted[N-i-1] != fixation_values_sorted[N-i-2]:
                    precs[correct_count] = float(correct_count)/fix_count
                    false_positives[correct_count] = float(false_positive_count)/false_count
                else:
                    precs[correct_count] = precs[correct_count - 1]
                    false_positives[correct_count] = false_positives[correct_count - 1]
        if not judd:
            if i == N-1 or image_sorted[N-i-1] != image_sorted[N-i-2]:
                precs[i+1] = float(correct_count)/fix_count
                false_positives[i+1] = float(false_positive_count)/false_count
            else:
                precs[i+1] = precs[i]
                false_positives[i+1] = false_positives[i]
        #print false_positives[i+1]
    precs[length-1] = 1.0
    false_positives[length-1] = 1.0
    aoc = np.trapz(precs, false_positives)
    return aoc, precs, false_positives


#Do not check for index errors
@cython.boundscheck(False)
#Do not enable negativ indices
@cython.wraparound(False)
#Use native c division
@cython.cdivision(True)
def general_roc(np.ndarray[double, ndim=1] positives, np.ndarray[double, ndim=1] negatives, int judd=0):
    """calculate ROC score for given values of positive and negative
    distribution"""
    cdef np.ndarray[double, ndim=1] sorted_positives = np.sort(positives)[::-1]
    cdef np.ndarray[double, ndim=1] sorted_negatives = np.sort(negatives)[::-1]
    cdef np.ndarray[double, ndim=1] all_values
    if judd == 0:
        all_values = np.hstack([positives, negatives])
        all_values = np.sort(all_values)[::-1]
    else:
        min_val = np.min([sorted_positives[len(positives)-1], sorted_negatives[len(negatives)-1]])
        max_val = np.max([sorted_positives[0], sorted_negatives[0]])+1
        all_values = np.hstack((max_val, positives, min_val))
        all_values = np.sort(all_values)[::-1]
    cdef np.ndarray[double, ndim=1] false_positive_rates = np.zeros(len(all_values)+1)
    cdef np.ndarray[double, ndim=1] hit_rates = np.zeros(len(all_values)+1)
    cdef int true_positive_count = 0
    cdef int false_positive_count = 0
    cdef int positive_count = len(positives)
    cdef int negative_count = len(negatives)
    cdef int i
    cdef double theta
    for i in range(len(all_values)):
        theta = all_values[i]
        while true_positive_count < positive_count and sorted_positives[true_positive_count] >= theta:
            true_positive_count += 1
        while false_positive_count < negative_count and sorted_negatives[false_positive_count] >= theta:
            false_positive_count += 1
        false_positive_rates[i+1] = float(false_positive_count) / negative_count
        hit_rates[i+1] = float(true_positive_count) / positive_count
    auc = np.trapz(hit_rates, false_positive_rates)
    return auc, hit_rates, false_positive_rates


#Do not check for index errors
@cython.boundscheck(False)
#Do not enable negativ indices
@cython.wraparound(False)
#Use native c division
@cython.cdivision(True)
def general_rocs_per_positive(np.ndarray[double, ndim=1] positives, np.ndarray[double, ndim=1] negatives):
    """calculate ROC scores for each positive against a list of negatives
    distribution. The mean over the result will equal the return value of `general_roc`."""
    cdef np.ndarray[double, ndim=1] sorted_positives = np.sort(positives)
    cdef np.ndarray[double, ndim=1] sorted_negatives = np.sort(negatives)
    cdef np.ndarray[long, ndim=1] sorted_inds = np.argsort(positives)

    cdef np.ndarray[double, ndim=1] results = np.empty(len(positives))
    cdef int true_positive_count = 0
    cdef int false_positive_count = 0
    cdef int true_negative_count = 0
    cdef int positive_count = len(positives)
    cdef int negative_count = len(negatives)
    cdef int i
    cdef double last_theta = -np.inf
    cdef double theta

    cdef int true_negatives_count = 0
    cdef int equal_count = 0
    for i in range(len(sorted_positives)):
        theta = sorted_positives[i]
        #print('theta', theta)
        if theta == last_theta:
            #print('same')
            results[sorted_inds[i]] = (1.0*true_negatives_count + 0.5*equal_count) / negative_count
            continue

        true_negatives_count = true_negatives_count + equal_count

        while true_negatives_count < negative_count and sorted_negatives[true_negatives_count] < theta:
            true_negatives_count += 1
            #print('.')
        equal_count = 0
        while true_negatives_count + equal_count < negative_count and sorted_negatives[true_negatives_count+equal_count] <= theta:
            equal_count += 1
            #print('=')
        results[sorted_inds[i]] = (1.0*true_negatives_count + 0.5*equal_count) / negative_count

        last_theta = theta
    return results



from __future__ import print_function, division, absolute_import

import os.path
import warnings

import numpy as np
from imageio import imread
from scipy.special import logsumexp
from scipy.io import loadmat
from tqdm import tqdm

from .models import Model
from .saliency_map_models import SaliencyMapModel
from .datasets import get_image_hash, FileStimuli
from .utils import get_minimal_unique_filenames


def export_model_to_hdf5(model, stimuli, filename, compression=9, overwrite=True):
    """Export pysaliency model predictions for stimuli into hdf5 file

    model: Model or SaliencyMapModel
    stimuli: instance of FileStimuli
    filename: where to save hdf5 file to
    compression: how much to compress the data
    overwrite: if False, an existing file will be appended to and
      if for some stimuli predictions already exist, they will be
      kept.
    """
    assert isinstance(stimuli, FileStimuli)

    names = get_minimal_unique_filenames(stimuli.filenames)

    import h5py

    if overwrite:
        mode = 'w'
    else:
        mode = 'a'

    with h5py.File(filename, mode=mode) as f:
        for k, s in enumerate(tqdm(stimuli)):
            if not overwrite and names[k] in f:
                print("Skipping already existing entry", names[k])
                continue
            if isinstance(model, SaliencyMapModel):
                smap = model.saliency_map(s)
            elif isinstance(model, Model):
                smap = model.log_density(s)
            else:
                raise TypeError(type(model))
            f.create_dataset(names[k], data=smap, compression=compression)


class SaliencyMapModelFromFiles(SaliencyMapModel):
    def __init__(self, stimuli, files, **kwargs):
        super(SaliencyMapModelFromFiles, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.stimulus_ids = list(stimuli.stimulus_ids)
        self.files = files
        assert(len(files) == len(stimuli))

    def _saliency_map(self, stimulus):
        filename = self._file_for_stimulus(stimulus)
        return self._load_file(filename)

    def _file_for_stimulus(self, stimulus):
        stimulus_id = get_image_hash(stimulus)

        try:
            stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)
        except IndexError:
            raise IndexError("Stimulus id '{}' not found in stimuli!".format(stimulus_id))

        return self.files[stimulus_index]

    def _load_file(self, filename):
        _, ext = os.path.splitext(filename)
        if ext.lower() in ['.png', '.jpg', '.jpeg', '.tiff']:
            return imread(filename).astype(float)
        elif ext.lower() == '.npy':
            return np.load(filename).astype(float)
        elif ext.lower() == '.mat':
            data = loadmat(filename)
            variables = [v for v in data if not v.startswith('__')]
            if len(variables) > 1:
                raise ValueError('{} contains more than one variable: {}'.format(filename, variables))
            elif len(variables) == 0:
                raise ValueError('{} contains no data'.format(filename))
            return data[variables[0]]
        else:
            raise ValueError('Unkown file type: {}'.format(ext))


class SaliencyMapModelFromDirectory(SaliencyMapModelFromFiles):
    def __init__(self, stimuli, directory, **kwargs):
        if not isinstance(stimuli, FileStimuli):
            raise TypeError('SaliencyMapModelFromDirectory works only with FileStimuli!')

        self.directory = directory
        files = os.listdir(directory)
        stems = [os.path.splitext(f)[0] for f in files]

        stimuli_files = [os.path.basename(f) for f in stimuli.filenames]
        stimuli_stems = [os.path.splitext(f)[0] for f in stimuli_files]

        assert set(stimuli_stems).issubset(stems)

        indices = [stems.index(f) for f in stimuli_stems]

        files = [os.path.join(directory, f) for f in files]
        files = [files[i] for i in indices]

        super(SaliencyMapModelFromDirectory, self).__init__(stimuli, files, **kwargs)


class SaliencyMapModelFromFile(SaliencyMapModel):
    """
    This class exposes a list of saliency maps stored in a .mat file
    as a pysaliency SaliencyMapModel. Especially, it can be used
    to import LSUN submissions into pysaliency.
    """
    def __init__(self, stimuli, filename, key='results', **kwargs):
        super(SaliencyMapModelFromFile, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.filename = filename
        _, ext = os.path.splitext(filename)
        if ext.lower() == '.mat':
            self.load_matlab(filename, key=key)
        else:
            raise ValueError('Unkown filetype', filename)

    def load_matlab(self, filename, key='results'):
        import hdf5storage
        data = hdf5storage.loadmat(filename)[key]
        if len(data.shape) == 2 and len(self.stimuli) > 1:
            if data.shape[0] == 1:
                data = data[0]
            elif data.shape[1] == 1:
                data = data[:, 0]
            else:
                raise ValueError('Data has wrong shape: {} (need 1xN, Nx1 or N)'.format(data.shape))
        if len(data.shape) > 2:
            raise ValueError('Data has wrong shape: {} (need 1xN, Nx1 or N)'.format(data.shape))
        expected_shape = (len(self.stimuli),)
        if not data.shape == expected_shape:
            raise ValueError('Wrong number of saliency maps! Expected {}, got {}'.format(expected_shape, data.shape))
        self._saliency_maps = [data[i] for i in range(len(self.stimuli))]

    def _saliency_map(self, stimulus):
        stimulus_id = get_image_hash(stimulus)
        stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)
        smap = self._saliency_maps[stimulus_index]
        if smap.shape != (stimulus.shape[0], stimulus.shape[1]):
            raise ValueError('Wrong shape!')
        return smap


class ModelFromDirectory(Model):
    def __init__(self, stimuli, directory, **kwargs):
        super(ModelFromDirectory, self).__init__(**kwargs)
        self.internal_model = SaliencyMapModelFromDirectory(stimuli, directory, caching=False)

    def _log_density(self, stimulus):
        smap = self.internal_model.saliency_map(stimulus)
        if not -0.01 <= logsumexp(smap) <= 0.01:
            raise ValueError('Not a correct log density!')
        return smap


class HDF5SaliencyMapModel(SaliencyMapModel):
    """ exposes a HDF5 file with saliency maps as pysaliency model

        The stimuli have to be of type `FileStimuli`. For each
        stimulus file, the model expects a dataset with the same
        name in the dataset.
    """
    def __init__(self, stimuli, filename, check_shape=True, **kwargs):
        super(HDF5SaliencyMapModel, self).__init__(**kwargs)
        assert isinstance(stimuli, FileStimuli)
        self.stimuli = stimuli
        self.filename = filename
        self.check_shape=check_shape
        import h5py
        self.hdf5_file = h5py.File(self.filename, 'r')

    def _saliency_map(self, stimulus):
        stimulus_id = get_image_hash(stimulus)
        stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)
        stimulus_filename = self.stimuli.filenames[stimulus_index]
        _, filename = os.path.split(stimulus_filename)
        smap = self.hdf5_file[filename][:]
        if not smap.shape == (stimulus.shape[0], stimulus.shape[1]):
            if self.check_shape:
                warnings.warn('Wrong shape for stimulus {}'.format(filename))
        return smap


class HDF5Model(Model):
    """ exposes a HDF5 file with log densities as pysaliency model.

        For more detail see HDF5SaliencyMapModel
    """
    def __init__(self, stimuli, filename, check_shape=True, **kwargs):
        super(HDF5Model, self).__init__(**kwargs)
        self.parent_model = HDF5SaliencyMapModel(stimuli = stimuli,
                                                 filename = filename,
                                                 caching=False,
                                                 check_shape=check_shape)

    def _log_density(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)
        if not -0.01 <= logsumexp(smap) <= 0.01:
            raise ValueError('Not a correct log density!')
        return smap



from __future__ import print_function, division, absolute_import, unicode_literals

from tqdm import tqdm

import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.special import logsumexp
import tensorflow as tf

from .models import sample_from_logdensity
from .saliency_map_models import SaliencyMapModel
from .tf_utils import gauss_blur


def sample_batch_fixations(log_density, fixations_per_image, batch_size, rst=None):
    xs, ys = sample_from_logdensity(log_density, fixations_per_image*batch_size, rst=rst)
    ns = np.repeat(np.arange(batch_size, dtype=int), repeats=fixations_per_image)

    return xs, ys, ns


def _eval_metric(log_density, test_samples, fn, seed=42, fixation_count=120, batch_size=50, verbose=True):
    values = []
    weights = []
    count = 0

    rst = np.random.RandomState(seed=seed)

    with tqdm(total=test_samples, leave=False, disable=not verbose) as t:
        while count < test_samples:
            this_count = min(batch_size, test_samples-count)
            xs, ys, ns = sample_batch_fixations(log_density, fixations_per_image=fixation_count, batch_size=this_count, rst=rst)

            values.append(fn(ns, ys, xs, this_count))
            weights.append(this_count)
            count += this_count
            t.update(this_count)
    weights = np.asarray(weights, dtype=np.float64) / np.sum(weights)
    return np.average(values, weights=weights)


def constrained_descent(opt, loss, params, constraint, learning_rate):
    """
    opt: e.g. tf.train.GradientDescent()
    """
    assert len(params) == 1
    constraint_grad, = tf.gradients(constraint, params)

    constraint_grad_norm = tf.reduce_sum(tf.square(constraint_grad))
    normed_constraint_grad = constraint_grad / constraint_grad_norm

    grads_and_vars = opt.compute_gradients(loss, params)

    grads0 = grads_and_vars[0][0]
    params0 = params[0]
    # first step: make sure we are not running into negative values
    max_allowed_grad = params0/learning_rate
    projected_grad1 = tf.reduce_min([grads0, max_allowed_grad], axis=0)
                #projected_grad1 = grads0

    # second step: Make sure that the gradient does not walk
    # out of the constraint
    projected_grad2 = projected_grad1 - tf.reduce_sum(projected_grad1*constraint_grad) * normed_constraint_grad

    #projected_grad = grads0 - tf.reduce_sum(grads0*constraint_grad) * normed_constraint_grad

    #param = params[0]
    #active_set = (param <= 0)
    #active_grad = (projected_grad > 0)  # we are _descending_, i.e. grad>0 means we are trying to go down
    #mask = tf.logical_not(tf.logical_and(active_set, active_grad))
    #projected_grad = projected_grad * tf.cast(mask, 'float')

    grads_and_vars = [(projected_grad2, grads_and_vars[0][1])]

    tf_train_op = opt.apply_gradients(grads_and_vars)

    return tf_train_op


def build_fixation_maps(Ns, Ys, Xs, batch_size, height, width, dtype=tf.float32):
    indices = tf.stack((Ns, Ys, Xs), axis=1)

    fixation_maps = tf.scatter_nd(indices, updates=tf.ones((tf.shape(indices)[0], ), dtype=dtype),
                                  shape=(batch_size, height, width))

    return fixation_maps


def tf_similarity(saliency_map, empirical_saliency_maps):
    normalized_empirical_saliency_maps = empirical_saliency_maps / tf.reduce_sum(empirical_saliency_maps, reduction_indices=[1, 2], keepdims=True)
    normalized_saliency_map = saliency_map / tf.reduce_sum(saliency_map)
    minimums = tf.minimum(normalized_empirical_saliency_maps, tf.expand_dims(normalized_saliency_map, 0))

    similarities = tf.reduce_sum(minimums, reduction_indices=[1, 2])

    return similarities


def build_similarity_graph(saliency_map, ns, ys, xs, batch_size, height, width, kernel_size, truncate_gaussian, dtype=tf.float32):
    window_radius = int(kernel_size*truncate_gaussian)

    fixation_maps = build_fixation_maps(ns, ys, xs, batch_size, height, width, dtype=dtype)

    empirical_saliency_maps = gauss_blur(
        tf.expand_dims(fixation_maps, axis=3),
        kernel_size, windowradius=window_radius,
        mode='ZERO'
    )[:, :, :, 0]

    similarities = tf_similarity(saliency_map, empirical_saliency_maps)

    return similarities


def maximize_expected_sim(log_density, kernel_size,
                          train_samples_per_epoch, val_samples,
                          train_seed=43, val_seed=42,
                          fixation_count=100, batch_size=50,
                          max_batch_size=None,
                          verbose=True, session_config=None,
                          initial_learning_rate=1e-7,
                          backlook=1, min_iter=0, max_iter=1000,
                          truncate_gaussian=3,
                          learning_rate_decay_samples=None,
                          initial_saliency_map=None,
                          learning_rate_decay_scheme=None,
                          learning_rate_decay_ratio = 0.333333333,
                          minimum_learning_rate = 1e-11):
    """
       max_batch_size: maximum possible batch size to be used in validation
       learning rate decay samples: how often to decay the learning rate (using 1/k)

       learning_rate_decay_scheme: how to decay the learning rate:
           - None, "1/k": 1/k scheme
           - "validation_loss": if validation loss not better for last backlook
           steps

        learning_rate_decay_ratio: how much to decay learning rate if `learning_rate_decay_scheme` == 'validation_loss'
        minimum_learning_rate: stop optimization if learning rate would drop below this rate if using validation loss decay scheme

    """

    if max_batch_size is None:
        max_batch_size = batch_size

    if learning_rate_decay_scheme is None:
        learning_rate_decay_scheme = '1/k'

    if learning_rate_decay_samples is None:
        learning_rate_decay_samples = train_samples_per_epoch

    log_density_sum = logsumexp(log_density)
    if not -0.001 < log_density_sum < 0.001:
        raise ValueError("Log density not normalized! LogSumExp={}".format(log_density_sum))

    if initial_saliency_map is None:
        initial_value = gaussian_filter(np.exp(log_density), kernel_size, mode='constant')
    else:
        initial_value = initial_saliency_map

    if initial_value.min() < 0:
        initial_value -= initial_value.min()

    initial_value /= initial_value.sum()

    graph = tf.Graph()
    dtype = tf.float32

    height, width = log_density.shape

    with graph.as_default():
        # setting up
        saliency_map = tf.get_variable('saliency_map', shape=log_density.shape, dtype=dtype)

        Ns = tf.placeholder(tf.int32, shape=(None, ), name='ns')
        Ys = tf.placeholder(tf.int32, shape=(None, ), name='ys')
        Xs = tf.placeholder(tf.int32, shape=(None, ), name='xs')
        BatchSize = tf.placeholder(tf.int32, shape=(), name='batch_size')

        similarities = build_similarity_graph(saliency_map, Ns, Ys, Xs, BatchSize, height, width, kernel_size, truncate_gaussian, dtype=dtype)

        similarity = tf.reduce_mean(similarities)

        loss = -similarity

        # constraints

        constraint = tf.reduce_sum(saliency_map) - 1.0

        # training

        learning_rate = tf.Variable(1.0, dtype=dtype)
        opt = tf.train.GradientDescentOptimizer(learning_rate)

        train_op = constrained_descent(opt, loss, [saliency_map], constraint, learning_rate)

        intermediate = saliency_map * tf.cast((saliency_map >= 0), 'float')
        normalized_saliency_map = intermediate / tf.reduce_sum(intermediate)
        normalize_op = tf.assign(saliency_map, normalized_saliency_map)

    #print("starting session")
    with tf.Session(graph=graph, config=session_config) as session:

        def _val_loss(ns, ys, xs, batch_size):
            #print("running", end='', flush=True)
            ret = session.run(loss, {Ns: ns, Ys: ys, Xs: xs, BatchSize: batch_size})
            #print("done")
            return ret

        def val_loss():
            return _eval_metric(log_density, val_samples, _val_loss, seed=val_seed,
                                fixation_count=fixation_count, batch_size=max_batch_size, verbose=False)

        session.run(tf.global_variables_initializer())
        #initial_value = gaussian_filter(density, kernel_size, mode='nearest')
        session.run(tf.assign(saliency_map, initial_value))
        session.run(tf.assign(learning_rate, initial_learning_rate))

        total_samples = 0
        decay_step = 0

        #print('starting val')

        val_scores = [val_loss()]
        learning_rate_relevant_scores = list(val_scores)
        train_rst = np.random.RandomState(seed=train_seed)
        #print('starting train')
        with tqdm(disable=not verbose) as outer_t:

            def general_termination_condition():
                return len(val_scores) - 1 >= max_iter

            def termination_1overk():
                return not (np.argmin(val_scores) >= len(val_scores) - backlook)

            def termination_validation():
                return session.run(learning_rate) < minimum_learning_rate

            def termination_condition():
                if len(val_scores) < min_iter:
                    return False
                cond = general_termination_condition()
                if learning_rate_decay_scheme == '1/k':
                    cond = cond or termination_1overk()
                elif learning_rate_decay_scheme == 'validation_loss':
                    cond = cond or termination_validation()

                return cond

            #while (len(val_scores) - 1 < max_iter) and (len(val_scores) < min_iter or np.argmin(val_scores) >= len(val_scores) - backlook):
            while not termination_condition():
                count = 0
                with tqdm(total=train_samples_per_epoch, leave=False, disable=True) as t:
                    while count < train_samples_per_epoch:
                        this_count = min(batch_size, train_samples_per_epoch-count)

                        xs, ys, ns = sample_batch_fixations(log_density, fixations_per_image=fixation_count, batch_size=this_count, rst=train_rst)
                        session.run(train_op, {Ns: ns, Ys: ys, Xs: xs, BatchSize: this_count})
                        session.run(normalize_op)

                        count += this_count
                        total_samples += this_count

                        if learning_rate_decay_scheme == '1/k':
                            if total_samples >= (decay_step+1)*learning_rate_decay_samples:
                                decay_step += 1
                                session.run(tf.assign(learning_rate, initial_learning_rate/decay_step))

                        t.update(this_count)
                val_scores.append(val_loss())
                learning_rate_relevant_scores.append(val_scores[-1])

                if np.argmin(learning_rate_relevant_scores) < len(learning_rate_relevant_scores) - backlook:
                    old_learning_rate = session.run(learning_rate)
                    #print("Old Learning Rate", old_learning_rate, type(old_learning_rate))
                    #print("Decay", learning_rate_decay_ratio)
                    new_learning_rate = old_learning_rate*learning_rate_decay_ratio
                    session.run(tf.assign(learning_rate, new_learning_rate))
                    #print("Decaying learning_rate to", new_learning_rate)
                    learning_rate_relevant_scores = [learning_rate_relevant_scores[-1]]

                score1, score2 = val_scores[-2:]
                last_min = len(val_scores) - np.argmin(val_scores) - 1
                outer_t.set_description('{:.05f}, diff {:.02e}, best val {} steps ago, lr {:.02e}'.format(score2, score2-score1, last_min, session.run(learning_rate)))
                outer_t.update(1)

        return session.run(saliency_map), val_scores[-1]


class SIMSaliencyMapModel(SaliencyMapModel):
    def __init__(self, parent_model,
                 kernel_size,
                 train_samples_per_epoch=1000, val_samples=1000,
                 train_seed=43, val_seed=42,
                 fixation_count=100, batch_size=50,
                 max_batch_size=None,
                 initial_learning_rate=1e-7,
                 backlook=1,
                 min_iter=0,
                 max_iter=1000,
                 truncate_gaussian=3,
                 learning_rate_decay_samples=None,
                 learning_rate_decay_scheme=None,
                 learning_rate_decay_ratio = 0.333333333,
                 minimum_learning_rate = 1e-11,
                 initial_model=None,
                 verbose=True,
                 session_config=None,
                 **kwargs
                 ):
        super(SIMSaliencyMapModel, self).__init__(**kwargs)
        self.parent_model = parent_model

        self.kernel_size = kernel_size
        self.train_samples_per_epoch = train_samples_per_epoch
        self.val_samples = val_samples
        self.train_seed = train_seed
        self.val_seed = val_seed
        self.fixation_count = fixation_count
        self.batch_size = batch_size
        self.max_batch_size = max_batch_size
        self.initial_learning_rate = initial_learning_rate
        self.backlook = backlook
        self.min_iter = min_iter
        self.max_iter = max_iter
        self.truncate_gaussian = truncate_gaussian
        self.learning_rate_decay_samples = learning_rate_decay_samples
        self.learning_rate_decay_scheme = learning_rate_decay_scheme
        self.learning_rate_decay_ratio = learning_rate_decay_ratio
        self.minimum_learning_rate = minimum_learning_rate
        self.initial_model = initial_model
        self.verbose = verbose
        self.session_config = session_config

    def _saliency_map(self, stimulus):
        log_density = self.parent_model.log_density(stimulus)

        if self.initial_model:
            initial_saliency_map = self.initial_model.saliency_map(stimulus)
        else:
            initial_saliency_map = None

        saliency_map, val_scores = maximize_expected_sim(
            log_density,
            kernel_size=self.kernel_size,
            train_samples_per_epoch=self.train_samples_per_epoch,
            val_samples=self.val_samples,
            train_seed=self.train_seed,
            val_seed=self.val_seed,
            fixation_count=self.fixation_count,
            batch_size=self.batch_size,
            max_batch_size=self.max_batch_size,
            verbose=self.verbose,
            session_config=self.session_config,
            initial_learning_rate=self.initial_learning_rate,
            backlook=self.backlook,
            min_iter=self.min_iter,
            max_iter=self.max_iter,
            truncate_gaussian=self.truncate_gaussian,
            learning_rate_decay_samples=self.learning_rate_decay_samples,
            initial_saliency_map=initial_saliency_map,
            learning_rate_decay_scheme=self.learning_rate_decay_scheme,
            learning_rate_decay_ratio=self.learning_rate_decay_ratio,
            minimum_learning_rate=self.minimum_learning_rate
        )
        return saliency_map



from __future__ import print_function, unicode_literals, division, absolute_import

import numba
import numpy as np
from scipy.special import logsumexp
from scipy.ndimage.filters import gaussian_filter

from sklearn.neighbors import KernelDensity
from sklearn.model_selection import BaseCrossValidator
from sklearn.base import BaseEstimator

from tqdm import tqdm

from .precomputed_models import get_image_hash
from .roc import general_roc
from .numba_utils import fill_fixation_map
from . import Model, UniformModel


@numba.jit(nopython=True)
def _normalize_fixations(orig_xs, orig_ys, orig_ns, sizes, new_xs, new_ys, real_widths, real_heights):
    for i in range(len(orig_xs)):
        height, width = sizes[orig_ns[i]]
        new_xs[i] = orig_xs[i] / width
        new_ys[i] = orig_ys[i] / height
        real_widths[i] = width
        real_heights[i] = height


def normalize_fixations(stimuli, fixations, keep_aspect=False, add_shape=False, verbose=True):
    sizes = np.array(stimuli.sizes)

    xs = np.empty(len(fixations.x))
    ys = np.empty(len(fixations.x))
    widths = np.empty(len(fixations.x))
    heights = np.empty(len(fixations.x))

    _normalize_fixations(fixations.x, fixations.y, fixations.n, sizes,
                         xs, ys, widths, heights)

    real_widths = widths.copy()
    real_heights = heights.copy()

    if keep_aspect:
        max_size = np.max([widths, heights], axis=0)
        widths = max_size
        heights = max_size

    xs = fixations.x / widths
    ys = fixations.y / heights

    real_widths /= widths
    real_heights /= heights

    if add_shape:
        return xs, ys, real_widths, real_heights

    return xs, ys


def fixations_to_scikit_learn(fixations, normalize=None, keep_aspect=False, add_shape=False,
                              add_stimulus_number=False,
                              add_fixation_number=False,
                              verbose=True):
    if normalize is None:
        xs = fixations.x
        ys = fixations.y
        data = [xs, ys]
        if add_shape:
            raise NotImplementedError()
    else:
        data = normalize_fixations(normalize, fixations, keep_aspect=keep_aspect, add_shape=add_shape,
                                   verbose=verbose)
    if add_stimulus_number:
        data = list(data) + [fixations.n]
    if add_fixation_number:
        data = list(data) + [np.arange(len(fixations.n))]
    return np.vstack(data).T.copy()


class ScikitLearnImageCrossValidationGenerator(object):
    def __init__(self, stimuli, fixations):
        self.stimuli = stimuli
        self.fixations = fixations

    def __iter__(self):
        for n in range(len(self.stimuli)):
            inds = self.fixations.n == n
            if inds.sum():
                yield ~inds, inds

    def __len__(self):
        return len(self.stimuli)


class ScikitLearnImageSubjectCrossValidationGenerator(object):
    def __init__(self, stimuli, fixations):
        self.stimuli = stimuli
        self.fixations = fixations

    def __iter__(self):
        for n in range(len(self.stimuli)):
            for s in range(self.fixations.subject_count):
                image_inds = self.fixations.n == n
                subject_inds = self.fixations.subjects == s
                train_inds, test_inds = image_inds & ~subject_inds, image_inds & subject_inds
                if test_inds.sum() == 0 or train_inds.sum() == 0:
                    #print("Skipping")
                    continue
                yield train_inds, test_inds

    def __len__(self):
        return len(set(zip(self.fixations.n, self.fixations.subjects)))


class ScikitLearnWithinImageCrossValidationGenerator(object):
    def __init__(self, stimuli, fixations, chunks_per_image=10, random_seed=42):
        self.stimuli = stimuli
        self.fixations = fixations
        self.chunks_per_image = chunks_per_image
        self.rng = np.random.RandomState(seed=random_seed)

    def __iter__(self):
        for n in range(len(self.stimuli)):
            image_inds = self.fixations.n == n
            _image_inds = np.nonzero(image_inds)[0]
            self.rng.shuffle(_image_inds)
            chunks = np.array_split(_image_inds, self.chunks_per_image)
            for chunk in chunks:
                if not len(chunk):
                    continue
                test_inds = np.zeros_like(self.fixations.n)
                test_inds[chunk] = 1
                test_inds = test_inds > 0.5
                train_inds = image_inds & ~test_inds
                yield train_inds, test_inds

    def __len__(self):
        #counts = 0
        #for n in range(len(self.stimuli)):

        return len(self.stimuli)*self.chunks_per_image


class RegularizedKernelDensityEstimator(BaseEstimator):
    def __init__(self, bandwidth=1.0, regularization = 1.0e-5):
        self.bandwidth = bandwidth
        self.regularization = regularization

    def setup(self):
        self.kde = KernelDensity(kernel='gaussian', bandwidth=self.bandwidth)

        height, width = self.shape
        self.uniform_density = -np.log(width*height)

        self.kde_constant = np.log(1-self.regularization)
        self.uniform_constant = np.log(self.regularization)

    def fit(self, X):
        self.shape = X[0, 2:4]
        self.setup()
        self.kde.fit(X[:, 0:2])
        return self

    def score_samples(self, X):
        kde_logliks = self.kde.score_samples(X[:, :2])

        logliks = np.logaddexp(
            self.kde_constant + kde_logliks,
            self.uniform_constant + self.uniform_density
        )
        return logliks

    def score(self, X):
        return np.sum(self.score_samples(X))


class MixtureKernelDensityEstimator(BaseEstimator):
    def __init__(self, bandwidth=1.0, regularization = 1.0e-5, regularizing_log_likelihoods=None):
        self.bandwidth = bandwidth
        self.regularization = regularization
        #self.regularizer_model = regularizer_model
        ##self.stimuli = stimuli
        self.regularizing_log_likelihoods = regularizing_log_likelihoods

    def setup(self):
        self.kde = KernelDensity(kernel='gaussian', bandwidth=self.bandwidth)

        self.kde_constant = np.log(1-self.regularization)
        self.uniform_constant = np.log(self.regularization)

    def fit(self, X):
        assert X.shape[1] == 3

        self.setup()
        self.kde.fit(X[:, 0:2])
        return self

    def score_samples(self, X):
        assert X.shape[1] == 3

        kde_logliks = self.kde.score_samples(X[:, :2])
        fix_ns = X[:, 2].astype(int)
        fix_lls = self.regularizing_log_likelihoods[fix_ns]

        logliks = np.logaddexp(
            self.kde_constant + kde_logliks,
            self.uniform_constant + fix_lls
        )
        return logliks

    def score(self, X):
        return np.sum(self.score_samples(X))


class AUCKernelDensityEstimator(BaseEstimator):
    def __init__(self, nonfixations, bandwidth=1.0):
        self.bandwidth = bandwidth
        self.nonfixations = nonfixations

    def setup(self):
        self.kde = KernelDensity(kernel='gaussian', bandwidth=self.bandwidth)

    def fit(self, X):
        self.setup()
        self.kde.fit(X)
        self.nonfixation_values = self.kde.score_samples(self.nonfixations)
        return self

    def score_samples(self, X):
        pos_logliks = self.kde.score_samples(X)
        neg_logliks = self.nonfixation_values

        aucs = [general_roc(np.array([p]), neg_logliks)[0] for p in pos_logliks]

        return aucs

    def score(self, X):
        return np.sum(self.score_samples(X))


class GoldModel(Model):
    def __init__(self, stimuli, fixations, bandwidth, eps = 1e-20, keep_aspect=False, verbose=False, **kwargs):
        super(GoldModel, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.fixations = fixations
        self.bandwidth = bandwidth
        self.eps = eps
        self.keep_aspect = keep_aspect
        self.xs, self.ys = normalize_fixations(stimuli, fixations, keep_aspect=self.keep_aspect, verbose=verbose)
        self.shape_cache = {}

    def _log_density(self, stimulus):
        shape = stimulus.shape[0], stimulus.shape[1]

        stimulus_id = get_image_hash(stimulus)
        stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)

        #fixations = self.fixations[self.fixations.n == stimulus_index]
        inds = self.fixations.n == stimulus_index

        if not inds.sum():
            return UniformModel().log_density(stimulus)

        ZZ = np.zeros(shape)
        if self.keep_aspect:
            height, width = shape
            max_size = max(width, height)
            x_factor = max_size
            y_factor = max_size
        else:
            x_factor = shape[1]
            y_factor = shape[0]
        _fixations = np.array([self.ys[inds]*y_factor, self.xs[inds]*x_factor]).T
        fill_fixation_map(ZZ, _fixations)
        ZZ = gaussian_filter(ZZ, [self.bandwidth*y_factor, self.bandwidth*x_factor])
        ZZ *= (1-self.eps)
        ZZ += self.eps * 1.0/(shape[0]*shape[1])
        ZZ = np.log(ZZ)

        ZZ -= logsumexp(ZZ)
        #ZZ -= np.log(np.exp(ZZ).sum())

        return ZZ



class KDEGoldModel(Model):
    def __init__(self, stimuli, fixations, bandwidth, eps = 1e-20, keep_aspect=False, verbose=False, **kwargs):
        super(KDEGoldModel, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.fixations = fixations
        self.bandwidth = bandwidth
        self.eps = eps
        self.keep_aspect = keep_aspect
        self.xs, self.ys = normalize_fixations(stimuli, fixations, keep_aspect=self.keep_aspect, verbose=verbose)
        self.shape_cache = {}

    def _log_density(self, stimulus):
        shape = stimulus.shape[0], stimulus.shape[1]

        stimulus_id = get_image_hash(stimulus)
        stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)

        #fixations = self.fixations[self.fixations.n == stimulus_index]
        inds = self.fixations.n == stimulus_index

        if not inds.sum():
            return UniformModel().log_density(stimulus)

        X = fixations_to_scikit_learn(
            self.fixations[inds], normalize=self.stimuli,
            keep_aspect=self.keep_aspect, add_shape=False, verbose=False)
        kde = KernelDensity(bandwidth=self.bandwidth).fit(X)

        height, width = shape
        if self.keep_aspect:
            max_size = max(height, width)
            rel_height = height / max_size
            rel_width = width / max_size
        else:
            rel_height = 1.0
            rel_width = 1.0

        # calculate the KDE score at the middle of each pixel:
        # for a width of 10 pixels, we are going to calculate at
        # 0.5, 1.5, ..., 9.5, since e.g. fixations with x coordinate between 0.0 and 1.0
        # will be evaluated at pixel index 0.
        xs = np.linspace(0, rel_width, num=width, endpoint=False)+0.5*rel_width/width
        ys = np.linspace(0, rel_height, num=height, endpoint=False)+0.5*rel_height/height
        XX, YY = np.meshgrid(xs, ys)
        scores = kde.score_samples(np.column_stack((XX.flatten(), YY.flatten()))).reshape(XX.shape)
        scores -= logsumexp(scores)
        ZZ = scores

        if self.eps:
            ZZ = np.logaddexp(
                np.log(1-self.eps)+scores,
                np.log(self.eps)-np.log(height*width)
            )

        ZZ -= logsumexp(ZZ)

        return ZZ



class CrossvalidatedBaselineModel(Model):
    def __init__(self, stimuli, fixations, bandwidth, eps = 1e-20, **kwargs):
        super(CrossvalidatedBaselineModel, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.fixations = fixations
        self.bandwidth = bandwidth
        self.eps = eps
        self.xs, self.ys = normalize_fixations(stimuli, fixations)
        #self.kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(np.vstack([self.xs, self.ys]).T)
        self.shape_cache = {}

    def _log_density(self, stimulus):
        shape = stimulus.shape[0], stimulus.shape[1]

        stimulus_id = get_image_hash(stimulus)
        stimulus_index = self.stimuli.stimulus_ids.index(stimulus_id)

        #fixations = self.fixations[self.fixations.n == stimulus_index]
        inds = self.fixations.n != stimulus_index

        ZZ = np.zeros(shape)

        _fixations = np.array([self.ys[inds]*shape[0], self.xs[inds]*shape[1]]).T
        fill_fixation_map(ZZ, _fixations)
        ZZ = gaussian_filter(ZZ, [self.bandwidth*shape[0], self.bandwidth*shape[1]])
        ZZ *= (1-self.eps)
        ZZ += self.eps * 1.0/(shape[0]*shape[1])
        ZZ = np.log(ZZ)

        ZZ -= logsumexp(ZZ)
        #ZZ -= np.log(np.exp(ZZ).sum())

        return ZZ


class BaselineModel(Model):
    def __init__(self, stimuli, fixations, bandwidth, eps = 1e-20, keep_aspect=False, **kwargs):
        super(BaselineModel, self).__init__(**kwargs)
        self.stimuli = stimuli
        self.fixations = fixations
        self.bandwidth = bandwidth
        self.eps = eps
        self.keep_aspect = keep_aspect
        self.xs, self.ys = normalize_fixations(stimuli, fixations, keep_aspect=keep_aspect)
        self.shape_cache = {}

    def _log_density(self, stimulus):
        shape = stimulus.shape[0], stimulus.shape[1]
        if shape not in self.shape_cache:
            ZZ = np.zeros(shape)
            height, width = shape
            if self.keep_aspect:
                max_size = max(height, width)
                y_factor = max_size
                x_factor = max_size
            else:
                y_factor = height
                x_factor = width
            _fixations = np.array([self.ys*y_factor, self.xs*x_factor]).T
            fill_fixation_map(ZZ, _fixations)
            ZZ = gaussian_filter(ZZ, [self.bandwidth*y_factor, self.bandwidth*x_factor])
            ZZ *= (1-self.eps)
            ZZ += self.eps * 1.0/(shape[0]*shape[1])
            ZZ = np.log(ZZ)

            ZZ -= logsumexp(ZZ)
            self.shape_cache[shape] = ZZ

        return self.shape_cache[shape]



from __future__ import absolute_import, print_function, division, unicode_literals

import os
import tempfile
import zipfile
import tarfile
from pkg_resources import resource_string,  resource_listdir

from boltons.fileutils import mkdir_p
import numpy as np
from scipy.ndimage import zoom

from .utils import TemporaryDirectory, download_and_check, run_matlab_cmd
from .quilt import QuiltSeries
from .saliency_map_models import MatlabSaliencyMapModel, SaliencyMapModel


def write_file(filename, contents):
    """Write contents to file and close file savely"""
    with open(filename, 'wb') as f:
        f.write(contents)


def extract_zipfile(filename, extract_to):
    if zipfile.is_zipfile(filename):
        z = zipfile.ZipFile(filename)
        #os.makedirs(extract_to)
        z.extractall(extract_to)
    elif tarfile.is_tarfile(filename):
        t = tarfile.open(filename)
        t.extractall(extract_to)
    else:
        raise ValueError('Unkown archive type', filename)


def unpack_directory(package, resource_name, location):
    files = resource_listdir(package, resource_name)
    for file in files:
        write_file(os.path.join(location, file),
                   resource_string(package, os.path.join(resource_name, file)))


def apply_quilt(source_location, package, resource_name, patch_directory, verbose=True):
    """Apply quilt series from package data to source code"""
    os.makedirs(patch_directory)
    unpack_directory(package, resource_name, patch_directory)
    series = QuiltSeries(patch_directory)
    series.apply(source_location, verbose=verbose)


def download_extract_patch(url, hash, location, location_in_archive=True, patches=None):
    """Download, extract and maybe patch code"""
    with TemporaryDirectory() as temp_dir:
        if not os.path.isdir(temp_dir):
            os.makedirs(temp_dir)
        archive_name = os.path.basename(url)
        download_and_check(url,
                           os.path.join(temp_dir, archive_name),
                           hash)

        if location_in_archive:
            target = os.path.dirname(os.path.normpath(location))
        else:
            target = location
        extract_zipfile(os.path.join(temp_dir, archive_name),
                        target)

    if patches:
        parent_directory = os.path.dirname(os.path.normpath(location))
        patch_directory = os.path.join(parent_directory, os.path.basename(patches))
        apply_quilt(location, __name__,  os.path.join('scripts', 'models', patches), patch_directory)


class ExternalModelMixin(object):
    """
    Download and cache necessary files.

    If the location is None, a temporary directory will be used.
    If the location is not None, the data will be stored in a
    subdirectory of location named after `__modelname`. If this
    sub directory already exists, the initialization will
    not be run.

    After running `setup()`, the actual location will be
    stored in `self.location`.

    To make use of this Mixin, overwrite `_setup()`
    and run `setup(location)`.
    """
    def setup(self, location, *args, **kwargs):
        if location is None:
            self.location = tempfile.mkdtemp()
            self._setup(*args, **kwargs)
        else:
            self.location = os.path.join(location, self.__modelname__)
            if not os.path.exists(self.location):
                self._setup(*args, **kwargs)

    def _setup(self, *args, **kwargs):
        raise NotImplementedError()


class AIM(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    The saliency model "Attention based on Information Maximization" (AIM)
    by Bruce and Tsotso. The original matlab code is used.

    .. seealso::
        Bruce, Tsotso: Saliency, attention, and visual search: An information theoretic approach. 2009.

        http://www.cs.umanitoba.ca/~bruce/datacode.html
    """
    __modelname__ = 'AIM'

    def __init__(self, filters='31jade950', convolve=True, location=None, **kwargs):
        self.setup(location)
        super(AIM, self).__init__(os.path.join(self.location, 'AIM_wrapper.m'), only_color_stimuli=True, **kwargs)
        self.filters = filters
        self.convolve = convolve

    def matlab_command(self, stimulus):
        return "AIM_wrapper('{{stimulus}}', '{{saliency_map}}', {}, '{}');".format(1 if self.convolve else 0,
                                                                                   self.filters)

    def _setup(self):
        with TemporaryDirectory() as temp_dir:
            if not os.path.isdir(temp_dir):
                os.makedirs(self.location)
            download_and_check('http://www.cs.umanitoba.ca/~bruce/AIM.zip',
                               os.path.join(temp_dir, 'AIM.zip'),
                               '6d52bc2c0cb15bc186d3d6de32751351')

            z = zipfile.ZipFile(os.path.join(temp_dir, 'AIM.zip'))
            namelist = z.namelist()
            namelist = [n for n in namelist if n.endswith('.m') or n.endswith('.mat')]

            z.extractall(self.location, namelist)
            with open(os.path.join(self.location, 'AIM_wrapper.m'), 'wb') as f:
                f.write(resource_string(__name__, 'scripts/models/AIM_wrapper.m'))


class SUN(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    The saliency model "Saliency using natural image statistics" (SUN)
    by Zhang et al. The original matlab code is used.

    .. note::
        Due to the patch based approach, the original code
        returns a saliency map that is smaller than the original
        stimulus. Here, in a wrapper script, we enlargen the
        saliency map to match the original stimulus by setting the
        boundary area to the minimal saliency value, as is
        done also by Bruce and Tsotso in AIM.

        As the original code can handle only color images,
        we convert grayscale images to color images by setting
        all color channels to the grayscale value.

        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Lingyun Zhang, Matthew H. Tong, Tim K. Marks, Honghao Shan, Garrison W. Cottrell. SUN: A Bayesian framework for saliency using natural statistics [JoV 2008]

        http://cseweb.ucsd.edu/~l6zhang/
    """
    __modelname__ = 'SUN'

    def __init__(self, scale=1.0, location=None, **kwargs):
        self.setup(location)
        super(SUN, self).__init__(os.path.join(self.location, 'SUN_wrapper.m'), only_color_stimuli=True, **kwargs)
        self.scale = scale

    def matlab_command(self, stimulus):
        return "SUN_wrapper('{{stimulus}}', '{{saliency_map}}', {});".format(self.scale)

    def _setup(self):
        with TemporaryDirectory() as temp_dir:
            if not os.path.isdir(temp_dir):
                os.makedirs(self.location)
            download_and_check('http://cseweb.ucsd.edu/~l6zhang/code/imagesaliency.zip',
                               os.path.join(temp_dir, 'SUN.zip'),
                               'df69e6c34b2e9e5ddd7a051d98e880d0')

            z = zipfile.ZipFile(os.path.join(temp_dir, 'SUN.zip'))
            namelist = z.namelist()
            namelist = [n for n in namelist if n.endswith('.m') or n.endswith('.mat')]

            z.extractall(self.location, namelist)
            with open(os.path.join(self.location, 'SUN_wrapper.m'), 'wb') as f:
                f.write(resource_string(__name__, 'scripts/models/SUN_wrapper.m'))
            with open(os.path.join(self.location, 'ensure_image_is_color_image.m'), 'wb') as f:
                f.write(resource_string(__name__, 'scripts/ensure_image_is_color_image.m'))


class ContextAwareSaliency(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    The saliency model "Context-Aware Saliency" by Goferman et. al.
    The original matlab code is used.

    .. seealso::
        Stas Goferman, Lihi Zelnik-Manor, Ayellet Tal. Context-Aware Saliency Detection [CVPR 2010] [PAMI 2012]

        http://webee.technion.ac.il/labs/cgm/Computer-Graphics-Multimedia/Software/Saliency/Saliency.html
    """
    __modelname__ = 'ContextAwareSaliency'

    def __init__(self, location=None, **kwargs):
        self.setup(location)
        super(ContextAwareSaliency, self).__init__(os.path.join(self.location, 'ContextAwareSaliency_wrapper.m'), **kwargs)

    def _setup(self):
        with TemporaryDirectory() as temp_dir:
            if not os.path.isdir(temp_dir):
                os.makedirs(self.location)
            download_and_check('http://webee.technion.ac.il/labs/cgm/Computer-Graphics-Multimedia/Software/Saliency/Saliency.zip',
                               os.path.join(temp_dir, 'Saliency.zip'),
                               'c3c6768ef26e95def76000f51e8aad7c')

            z = zipfile.ZipFile(os.path.join(temp_dir, 'Saliency.zip'))
            source_location = os.path.join(self.location, 'source')
            os.makedirs(source_location)
            z.extractall(source_location)

            with open(os.path.join(self.location, 'ContextAwareSaliency_wrapper.m'), 'wb') as f:
                f.write(resource_string(__name__, 'scripts/models/ContextAwareSaliency_wrapper.m'))


class BMS(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    "Boolean Map based Saliency (BMS)" by Zhang and Sclaroff.
    The original matlab code is used.

    .. note::
        The original code is slightly patched to work from other directories.
        Also, their code uses OpenCV for performance. The original compile script
        had hardcoded Windows paths for the location of OpenCV. The compile script
        is patched to use the path of OpenCV in Ubuntu. If you want to use the BMS
        model in other systems, you might have to adapt the compile script and run
        it yourself.

        In Debian-based linux systems like Ubuntu you can install all needed
        opencv-libraries with

            apt-get install libopencv-core-dev libopencv-highgui-dev libopencv-imgproc-dev libopencv-flann-dev libopencv-photo-dev libopencv-video-dev libopencv-features2d-dev libopencv-objdetect-dev libopencv-calib3d-dev libopencv-ml-dev libopencv-contrib-dev

    .. seealso::
        Jianming Zhang, Stan Sclaroff. Saliency detection: a boolean map approach [ICCV 2013]
        [http://cs-people.bu.edu/jmzhang/BMS/BMS_iccv13_preprint.pdf]

        http://cs-people.bu.edu/jmzhang/BMS/BMS.html
    """
    __modelname__ = 'BMS'

    def __init__(self, location=None, **kwargs):
        self.setup(location)
        super(BMS, self).__init__(os.path.join(self.location, 'BMS_wrapper.m'), **kwargs)

    def _setup(self):
        with TemporaryDirectory() as temp_dir:
            if not os.path.isdir(temp_dir):
                os.makedirs(self.location)
            download_and_check('http://cs-people.bu.edu/jmzhang/BMS/BMS-mex.zip',
                               os.path.join(temp_dir, 'BMS.zip'),
                               '056fa962993c3083f7977ee18432dd8c')

            source_location = os.path.join(self.location, 'source')
            extract_zipfile(os.path.join(temp_dir, 'BMS.zip'),
                            source_location)

            apply_quilt(source_location, __name__, os.path.join('scripts',
                                                                'models',
                                                                'BMS',
                                                                'patches'),
                        os.path.join(self.location, 'patches'))

            run_matlab_cmd('compile', cwd=os.path.join(source_location, 'mex'))

            with open(os.path.join(self.location, 'BMS_wrapper.m'), 'wb') as f:
                f.write(resource_string(__name__, 'scripts/models/BMS/BMS_wrapper.m'))


class GBVS(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    "Graph Based Visual Saliency (GBVS)" by Zhang and Sclaroff.
    The original matlab code is used.

    .. note::
        The original code is slightly patched to work from other directories.

        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Jonathan Harel, Christof Koch, Pietro Perona. Graph-Based Visual Saliency [NIPS 2006]

        http://www.vision.caltech.edu/~harel/share/gbvs.php
    """
    __modelname__ = 'GBVS'

    def __init__(self, location=None, **kwargs):
        """
        The parameter explanation have been adopted from the original code.

        Parameters
        ==========

        General
        -------

        @type  salmapmaxsize: integer, defaults to 32
        @param salmapmaxsize: size of calculated saliency maps (maximum dimension).
                              don't set this too high (e.g., >60).

        @type  blurfrac: float, defaults to 0.02
        @param blurfrac: final blur to apply to master saliency map
                         (in standard deviations of gaussian kernel,
                         expressed as fraction of image width).
                         Use value 0 to turn off this feature.

        Features
        --------

        @type  channels: string, defaults to `DIO`.
        @param channels: feature channels to use encoded as a string.
                         these are available:
                           C is for Color
                           I is for Intensity
                           O is for Orientation
                           R is for contRast
                           F is for Flicker
                           M is for Motion
                           D is for DKL Color (Derrington Krauskopf Lennie) ==
                             much better than C channel
                         e.g., 'IR' would be only intensity and
                               contrast, or
                         'CIO' would be only color,int.,ori. (standard)
                         'CIOR' uses col,int,ori, and contrast
        @type  colorWeight, intensityWeight, orientationWeight, contrastWeight, flickerWeight, motionWeight, dklcolorWeight: float, defaults to 1
        @param colorWeight, intensityWeight, orientationWeight, contrastWeight, flickerWeight, motionWeight, dklcolorWeight:
                  weights of feature channels (do not need to sum to 1).

        @type  gaborangles: list, None is the default: [0, 45, 90, 135]
        @param gaborangles: angles of gabor filters

        @type  contrastwidth: float, defaults to 0.1
        @param contrastwidth: fraction of image width = length of square side over which luminance variance is
                              computed for 'contrast' feature map. LARGER values will give SMOOTHER contrast maps.


        """
        self.setup(location)
        super(GBVS, self).__init__(os.path.join(self.location, 'GBVS_wrapper.m'), **kwargs)

    def _setup(self):
        source_location = os.path.join(self.location, 'gbvs')
        download_extract_patch('http://www.vision.caltech.edu/~harel/share/gbvs.zip',
                               'c5a86b9549c2c0bbd1b7f7e5b663b031',
                               source_location,
                               location_in_archive=True,
                               patches=os.path.join('GBVS', 'patches'))

        run_matlab_cmd("addpath('compile');gbvs_compile", cwd=source_location)

        with open(os.path.join(self.location, 'GBVS_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/GBVS/GBVS_wrapper.m'))


class GBVSIttiKoch(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    IttiKoch as implemented in "Graph Based Visual Saliency (GBVS)" by Zhang and Sclaroff.
    The original matlab code is used.

    .. note::
        The original code is slightly patched to work from other directories.

        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Jonathan Harel, Christof Koch, Pietro Perona. Graph-Based Visual Saliency [NIPS 2006]

        http://www.vision.caltech.edu/~harel/share/gbvs.php
    """
    __modelname__ = 'GBVSIttiKoch'

    def __init__(self, location=None, **kwargs):
        self.setup(location)
        super(GBVSIttiKoch, self).__init__(os.path.join(self.location, 'GBVSIttiKoch_wrapper.m'), **kwargs)

    def _setup(self):
        source_location = os.path.join(self.location, 'gbvs')
        download_extract_patch('http://www.vision.caltech.edu/~harel/share/gbvs.zip',
                               'c5a86b9549c2c0bbd1b7f7e5b663b031',
                               source_location,
                               location_in_archive=True,
                               patches=os.path.join('GBVS', 'patches'))

        run_matlab_cmd("addpath('compile');gbvs_compile", cwd=source_location)

        with open(os.path.join(self.location, 'GBVSIttiKoch_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/GBVS/GBVSIttiKoch_wrapper.m'))


class Judd(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    Judd model by Judd et al.
    The original matlab code is used.

    .. note::
        The original code is patched to work from other directories.

        The model makes use of the [SaliencyToolbox](http://www.saliencytoolbox.net/). Due
        to licence restrictions the Toolbox cannot be downloaded automatically. You have to
        download it yourself and provide the location of the zipfile via the
        `saliency_toolbox_archive`-keyword to the constructor.

        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to predict where humans look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    __modelname__ = 'Judd'

    def __init__(self, location=None, saliency_toolbox_archive=None, include_locations=None, library_locations=None, **kwargs):
        self.setup(location, saliency_toolbox_archive=saliency_toolbox_archive, include_locations=include_locations, library_locations=library_locations)
        super(Judd, self).__init__(os.path.join(self.location, 'Judd_wrapper.m'), only_color_stimuli=True, **kwargs)

    def _setup(self, saliency_toolbox_archive, include_locations, library_locations):
        if not saliency_toolbox_archive:
            raise Exception('You have to provide the zipfile containing the Itti and Koch Saliency Toolbox!')

        if include_locations is None:
            include_locations = ['/usr/include/opencv2']
        if library_locations is None:
            library_locations = ['/usr/lib/x86_64-linux-gnu/']

        source_location = os.path.join(self.location, 'source')
        print('Downloading Judd Model...')
        download_extract_patch('http://people.csail.mit.edu/tjudd/WherePeopleLook/Code/JuddSaliencyModel.zip',
                               '03e56c6f37c0ef3605c4c476f8a35b6b',
                               os.path.join(source_location, 'JuddSaliencyModel'),
                               location_in_archive=True,
                               patches=os.path.join('Judd', 'JuddSaliencyModel_patches'))

        print('Downloading matlabPyrTools...')
        download_extract_patch('http://www.cns.nyu.edu/pub/eero/matlabPyrTools.tar.gz',
                               'da53c02c843ed636bb35e20dac68a1b2',
                               os.path.join(source_location, 'matlabPyrTools'),
                               location_in_archive=True,
                               patches=None)

        print('Downloading voc')
        download_extract_patch('http://cs.brown.edu/~pff/latent-release3/voc-release3.1.tgz',
                               '20502f8a40f1122e00f81dcc0d11a843',
                               os.path.join(source_location, 'voc-release3.1'),
                               location_in_archive=True,
                               patches=os.path.join('Judd', 'voc_patches'))
        run_matlab_cmd("compile;quit;", cwd=os.path.join(source_location, 'voc-release3.1'))

        print('Extracting Saliency Toolbox')
        extract_zipfile(saliency_toolbox_archive, source_location)
        apply_quilt(os.path.join(source_location, 'SaliencyToolbox'),
                    __name__, os.path.join('scripts', 'models', 'Judd', 'SaliencyToolbox_patches'),
                    os.path.join(source_location, 'SaliencyToolbox_patches'))

        print('Downloading Viola Jones Face Detection')
        download_extract_patch('http://www.mathworks.com/matlabcentral/fileexchange/19912-open-cv-viola-jones-face-detection-in-matlab?download=true',
                               '26d3f6bc6641d959661afedadff8b479',
                               os.path.join(source_location, 'FaceDetect'),
                               location_in_archive=False,
                               patches=os.path.join('Judd', 'FaceDetect_patches'))

        includes = ' '.join(['-I{}'.format(location) for location in include_locations])
        libraries = ' '.join(['-L{}'.format(location) for location in library_locations])


        run_matlab_cmd("mex FaceDetect.cpp {includes} {libaries} -lopencv_core -lopencv_objdetect -lopencv_imgproc -lopencv_highgui -outdir .".format(
                           includes=includes,
                           libaries=libraries
                       ),
                       cwd=os.path.join(source_location, 'FaceDetect', 'src'))

        print('Downloading LabelMe Toolbox')
        download_extract_patch('http://labelme.csail.mit.edu/LabelMeToolbox/LabelMeToolbox.zip',
                               '66d982aae539149eff09ce26b7278f4d',
                               os.path.join(source_location, 'LabelMeToolbox'),
                               location_in_archive=True,
                               patches=None)

        with open(os.path.join(self.location, 'Judd_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/Judd/Judd_wrapper.m'))


class IttiKoch(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    Model by Itti and Koch.
    The original matlab code is used.

    .. note::
        The original code is patched to work from other directories.

        The model makes use of the [SaliencyToolbox](http://www.saliencytoolbox.net/). Due
        to licence restrictions the Toolbox cannot be downloaded automatically. You have to
        download it yourself and provide the location of the zipfile via the
        `saliency_toolbox_archive`-keyword to the constructor.

        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Saliency Toolbox by: Dirk Walther, Christof Koch. Modeling attention to salient proto-objects [Neural Networks 2006]

        http://www.saliencytoolbox.net/
    """
    __modelname__ = 'IttiKoch'

    def __init__(self, location=None, saliency_toolbox_archive=None, **kwargs):
        self.setup(location, saliency_toolbox_archive=saliency_toolbox_archive)
        super(IttiKoch, self).__init__(os.path.join(self.location, 'IttiKoch_wrapper.m'), only_color_stimuli=True, **kwargs)

    def _setup(self, saliency_toolbox_archive):
        if not saliency_toolbox_archive:
            raise Exception('You have to provide the zipfile containing the Itti and Koch Saliency Toolbox!')

        print('Extracting Saliency Toolbox')
        extract_zipfile(saliency_toolbox_archive, self.location)
        apply_quilt(os.path.join(self.location, 'SaliencyToolbox'),
                    __name__, os.path.join('scripts', 'models', 'Judd', 'SaliencyToolbox_patches'),
                    os.path.join(self.location, 'SaliencyToolbox_patches'))

        with open(os.path.join(self.location, 'IttiKoch_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/IttiKoch_wrapper.m'))


class RARE2012(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    RARE2012 by Margolin et al.
    The original matlab code is used.

    .. note::
        The original code is patched to work from other directories.

        The model makes use of the [SaliencyToolbox](http://www.saliencytoolbox.net/). Due
        to licence restrictions the Toolbox cannot be downloaded automatically. You have to
        download it yourself and provide the location of the zipfile via the
        `saliency_toolbox_archive`-keyword to the constructor.

        This model does not work with octave due to incompabilities
        of octave with matlab (RARE2012 comes as precompiled matlab file).
        This might change in the future.

    .. seealso::
        Nicolas Riche, Matei Mancas, Matthieu Duvinage, Makiese Mibulumukini,
        Bernard Gosselin, Thierry Dutoit. RARE2012: A multi-scale rarity-based
        saliency detection with its comparative statistical analysis
        [Signal Processing: Image Communication, 2013]

        http://www.tcts.fpms.ac.be/attention/?categorie16/what-and-why
    """
    __modelname__ = 'RARE2012'

    def __init__(self, location=None, **kwargs):
        self.setup(location)
        super(RARE2012, self).__init__(os.path.join(self.location, 'RARE2012_wrapper.m'), only_color_stimuli=True, **kwargs)

    def _setup(self):
        source_location = os.path.join(self.location, 'source')
        print('Downloading RARE2012 Model...')
        #download_extract_patch('http://www.tcts.fpms.ac.be/attention/data/documents/data/rare2012.zip',
        download_extract_patch('http://tcts.fpms.ac.be/attention/data/medias/documents/models/rare2012.zip',
                               '5a0a0de83e82b46fa70cea8b0a6bae55',
                               os.path.join(source_location, 'Rare2012'),
                               location_in_archive=True,
                               patches=None)

        print('Downloading simplegabortb-v1.0.0')
        download_extract_patch('http://www2.it.lut.fi/project/simplegabor/downloads/src/simplegabortb/simplegabortb-v1.0.0.tar.gz',
                               '92bc6ae7178a7b1301fad52489f5d677',
                               os.path.join(source_location, 'simplegabortb-v1.0.0'),
                               location_in_archive=True,
                               patches=None)

        with open(os.path.join(self.location, 'RARE2012_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/RARE2012_wrapper.m'))


class CovSal(ExternalModelMixin, MatlabSaliencyMapModel):
    """
    CovSal by Erdem and Erdem
    The original matlab code is used.

    .. note::
        This model does not work with octave due to incompabilities
        of octave with matlab. This might change in the future.

    .. seealso::
        Erkut Erdem, Aykut Erdem. Visual saliency estimation by nonlinearly integrating features using region covariances [JoV 2013]

        http://web.cs.hacettepe.edu.tr/~erkut/projects/CovSal/
    """
    __modelname__ = 'CovSal'

    def __init__(self, location=None, size=512, quantile=0.1, centerbias=True, modeltype='SigmaPoints', **kwargs):
        """
        Parameters
        ----------

        @type  size: int
        @param size: size of rescaled image

        @type  quantile: float
        @param quantile: parameter specifying the most similar regions in the neighborhood

        @type  centerbias: bool
        @param centerbias: True for center bias and False for no center bias

        @type  modeltype: string
        @param modeltype: 'CovariancesOnly' and 'SigmaPoints' to denote whether
                          first-order statistics will be incorporated or not

        """
        self.setup(location)
        super(CovSal, self).__init__(os.path.join(self.location, 'CovSal_wrapper.m'), only_color_stimuli=True, **kwargs)
        self.size = size
        self.quantile = quantile
        self.centerbias = centerbias
        if modeltype not in ['CovariancesOnly', 'SigmaPoints']:
            raise ValueError('Unkown modeltype', modeltype)
        self.modeltype = modeltype

    def matlab_command(self, stimulus):
        return "CovSal_wrapper('{{stimulus}}', '{{saliency_map}}', {}, {}, {}, '{}');".format(self.size,
                                                                                              self.quantile,
                                                                                              1 if self.centerbias else 0,
                                                                                              self.modeltype)

    def _setup(self):
        download_extract_patch('https://web.cs.hacettepe.edu.tr/~erkut/projects/CovSal/saliency.zip',
                               '5c1bcdce438a08051968b085f72b2fdc',
                               os.path.join(self.location, 'saliency'),
                               location_in_archive=True,
                               patches=None)

        with open(os.path.join(self.location, 'CovSal_wrapper.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/models/CovSal_wrapper.m'))



from __future__ import absolute_import, print_function, division, unicode_literals

import os
from glob import glob
import shutil
import h5py

from boltons.fileutils import mkdir_p
from boltons.iterutils import chunked
import numpy as np
from scipy.misc import logsumexp

from executor import execute
from tqdm import tqdm

from .external_models import ExternalModelMixin, download_extract_patch
from .models import GeneralModel
from .datasets import Stimuli, Fixations


def evaluate_scanpaths_hdf5(data, directory, scene_walker_directory, execute=execute, variables=None,
                            params=None, verbose=False, model='dynamic_model'):
    """evaluate log likelihood of scanpath on empirical_density given fixation data and size in degree

    data has to an array of {'saliency': saliency, deg_size: deg_size 'scanpaths': list of {'fixations', 'times'}}

    variables indicates which variables out of [fixations, likelihood, mapsA, mapsI, us] to return
    """

    if os.path.isdir(directory):
        for subdirectory in glob(os.path.join(directory, 'data')):
            print("Warning! Deleting subdirectory", subdirectory)
            shutil.rmtree(subdirectory)

        for subdirectory in glob(os.path.join(directory, 'image'+'[0-9]'*8)):
            print("Warning! Deleting subdirectory", subdirectory)
            shutil.rmtree(subdirectory)
        for hdf5file in glob(os.path.join(directory, '*.hdf5')):
            print("Warning! Deleting file", hdf5file)
            os.remove(hdf5file)

    mkdir_p(directory)

    if variables is None:
        variables = ['fixations', 'likelihood', 'mapsA', 'mapsI', 'us']

    if params:
        pars = np.array([
            params['omega'],
            params['rho'],
            params['sigmaAttention'],
            params['sigmaInhib'],
            params['gamma'],
            params['lambda'],
            params['inhibStrength'],
            params['zeta'],
        ])
    else:
        pars = np.array([])

    #print("Writing data")
    with h5py.File(os.path.join(directory, 'data.hdf5'), 'w') as hdf5file:
        hdf5file.create_dataset('image_count', data=len(data))
        for k, image_data in enumerate(tqdm(data, disable=not verbose)):

            image_group = hdf5file.create_group('image{k:08d}'.format(k=k))
            image_group.create_dataset('scanpath_count', data=len(image_data['scanpaths']))
            image_group.create_dataset('saliency', data=image_data['saliency'].astype(float).T)
            image_group.create_dataset('im_size', data=np.array(image_data['saliency'].shape).astype(float)[np.newaxis, :].T)
            image_group.create_dataset('deg_size', data=np.array(image_data['deg_size']).astype(float)[np.newaxis, :].T)
            image_group.create_dataset('pars', data=np.array(pars).astype(float))

            for i, scanpath_data in enumerate(image_data['scanpaths']):
                scanpath_group = image_group.create_group('scanpath{i:08d}'.format(i=i))

                scanpath_group.create_dataset('fixations', data=np.asarray(scanpath_data['fixations'], float).T)
                scanpath_group.create_dataset('times', data=np.asarray(scanpath_data['times'], float)[np.newaxis, :])

    full_scene_walker_directory = os.path.join(os.getcwd(), scene_walker_directory)

    save_cmds = []
    for variable_name in variables:

        source = variable_name

        if variable_name == 'last_us':
            source = 'us(:, :, end-1)'

        save_cmds.append("type_id = H5T.copy('H5T_IEEE_F64LE');")
        save_cmds.append("dims = fliplr(size({source}));".format(source=source))
        save_cmds.append("space_id = H5S.create_simple(length(dims), dims, dims);")
        save_cmds.append("dataset_id = H5D.create(output_scanpath_group, '{variable_name}', type_id, space_id, 'H5P_DEFAULT');".format(variable_name=variable_name))
        save_cmds.append("H5D.write(dataset_id, 'H5ML_DEFAULT', 'H5S_ALL', 'H5S_ALL', plist, {source});".format(source=source))

        save_cmds.append("H5S.close(space_id);")
        save_cmds.append("H5T.close(type_id);")
        save_cmds.append("H5D.close(dataset_id);")

    save_cmd = '\n                '.join(save_cmds)

    with open(os.path.join(directory, 'run.m'), 'w') as f:
        f.write("""
        addpath('{full_scene_walker_directory}');

        input_file_id = H5F.open('data.hdf5');
        output_file_id = H5F.create('results.hdf5');

        plist = 'H5P_DEFAULT';

        d_id = H5D.open(input_file_id, 'image_count');
        image_count = H5D.read(d_id);
        H5D.close(d_id);

        for image_no = 1 : image_count

            input_image_group = H5G.open(input_file_id, sprintf('/image%08d', image_no-1));
            output_image_group = H5G.create(output_file_id, sprintf('/image%08d', image_no-1), plist, plist, plist);

            d_id = H5D.open(input_image_group, 'im_size');
            im_size = H5D.read(d_id);
            H5D.close(d_id);

            d_id = H5D.open(input_image_group, 'deg_size');
            deg_size = H5D.read(d_id);
            H5D.close(d_id);

            d_id = H5D.open(input_image_group, 'saliency');
            saliency = H5D.read(d_id);
            H5D.close(d_id);

            d_id = H5D.open(input_image_group, 'pars');
            pars = H5D.read(d_id);
            H5D.close(d_id);

            d_id = H5D.open(input_image_group, 'scanpath_count');
            scanpath_count = H5D.read(d_id);
            H5D.close(d_id);

            fprintf('       ');

            for scanpath_no = 1 : scanpath_count


                input_scanpath_group = H5G.open(input_image_group, sprintf('/image%08d/scanpath%08d', image_no - 1, scanpath_no - 1));
                output_scanpath_group = H5G.create(output_image_group, sprintf('/image%08d/scanpath%08d', image_no - 1, scanpath_no - 1), plist, plist, plist);

                fprintf('\\rimage %d/%d scanpath %d/%d', image_no, image_count, scanpath_no, scanpath_count);

                d_id = H5D.open(input_scanpath_group, 'fixations');
                fixations = H5D.read(d_id);
                H5D.close(d_id);

                d_id = H5D.open(input_scanpath_group, 'times');
                times = H5D.read(d_id);
                H5D.close(d_id);

                [fixations, likelihood, mapsA, mapsI, us] = {model}(im_size, deg_size, saliency, fixations(1, :), times, fixations(2:end, :), pars);

                {save_cmd}

            end

            H5G.close(input_image_group);
            H5G.close(output_image_group);


        end

        fprintf('\\rDone.                                             \\n');

        H5F.close(input_file_id);
        H5F.close(output_file_id);

        """.format(
            full_scene_walker_directory=full_scene_walker_directory,
            save_cmd=save_cmd,
            model=model)
        )

    full_directory = os.path.join(os.getcwd(), directory)
    execute('matlab -nodesktop -r "try;run;catch exc;disp(getReport(exc));disp(\'__ERROR__\');exit(1);end;quit"',
            directory=full_directory)

    results = []

    #print("reading results")
    with h5py.File(os.path.join(directory, 'results.hdf5'), 'r') as results_file:
        for k, image_data in enumerate(tqdm(data, disable=not verbose)):
            results.append([])

            image_group = results_file['image{k:08d}'.format(k=k)]

            for i, scanpath_data in enumerate(image_data['scanpaths']):
                data = {}

                scanpath_group = image_group['scanpath{i:08d}'.format(i=i)]
                for variable in variables:
                    # matlab transposes the data before storing it in hdf5. We have to transpose it back.
                    data[variable] = scanpath_group[variable][...].T
                results[-1].append(data)

    for hdf5file in glob(os.path.join(directory, '*.hdf5')):
        os.remove(hdf5file)

    return results


class SceneWalker(ExternalModelMixin, GeneralModel):
    """The SceneWalker model

    .. seealso::
        TODO
    """

    __modelname__ = 'SceneWalker'

    def __init__(self,
                 base_model, pixel_per_dva, temporary_directory,
                 rho=1e30, omega=1.9298, gamma=44.780, lambda_=0.8115,
                 sigmaAttention = 5.9082*np.sqrt(0.8115),
                 sigmaInhib = 4.5531*np.sqrt(44.780),
                 inhibStrength = 0.3637,
                 zeta = 0.0722,
                 location=None,
                 execute=execute,
                 **kwargs):
        self.setup(location)
        super(SceneWalker, self).__init__(**kwargs)

        self.base_model = base_model
        self.pixel_per_dva = pixel_per_dva
        self.temporary_directory = temporary_directory
        self.execute = execute

        self.rho = rho
        self.omega = omega
        self.gamma = gamma
        self.lambda_ = lambda_
        self.sigmaAttention = sigmaAttention
        self.sigmaInhib = sigmaInhib
        self.inhibStrength = inhibStrength
        self.zeta = zeta

        self._model_function = 'dynamic_model'

    def _setup(self):
        mkdir_p(self.location)
        download_extract_patch('http://read.psych.uni-potsdam.de/attachments/article/174/PaperCode.zip',
                               'a5c6cf7f5b05281fced56a5841842f4c',
                               self.location,
                               location_in_archive=False,
                               patches=None)

    @property
    def params(self):
        return {
            'rho': self.rho,
            'omega': self.omega,
            'gamma': self.gamma,
            'lambda': self.lambda_,
            'sigmaAttention': self.sigmaAttention,
            'sigmaInhib': self.sigmaInhib,
            'inhibStrength': self.inhibStrength,
            'zeta': self.zeta
        }

    def _conditional_log_densities(self, stimuli, fixations, verbose=False):
        data = []
        data_locations = []
        stimulus_mapping = {}
        #print("preparing data")
        for i in tqdm(range(len(fixations.x)), disable=not verbose):

            n = fixations.n[i]
            if n not in stimulus_mapping:
                stimulus = stimuli[n]

                # SceneWalker expects the saliency map to be in "linear units", i.e. linear to probabilities
                saliency_map = np.exp(self.base_model.log_density(stimulus))

                size = np.array(stimulus.size, dtype=float)
                deg_size = size / self.pixel_per_dva

                data.append({
                    'saliency': saliency_map,
                    'deg_size': deg_size,
                    'scanpaths': []
                })
                stimulus_mapping[n] = len(data) - 1

            stimulus_data = data[stimulus_mapping[n]]

            length = fixations.lengths[i]
            x_hist = fixations.x_hist[i, :length]
            y_hist = fixations.y_hist[i, :length]
            t_hist = fixations.t_hist[i, :length]

            next_t = fixations.t[i]
            last_duration = next_t - t_hist[-1]

            _fixations = np.array([x_hist, y_hist]).T.astype(float)
            _fixations = np.vstack((_fixations, [0, 0]))  # we need to ask for another fixation to get the conditional probability
            _fixations /= self.pixel_per_dva

            times = np.array(t_hist)
            durations = np.diff(times)   # scene walker needs durations, not times
            durations = np.hstack((durations, [last_duration, 0.5]))  # we are missing the duration for the last fixation and need another fake fixation

            stimulus_data['scanpaths'].append({
                'fixations': _fixations,
                'times': durations
            })

            data_locations.append({'image_no': stimulus_mapping[n], 'scanpath_no': len(stimulus_data['scanpaths']) - 1})

        results = evaluate_scanpaths_hdf5(
            data = data,
            directory = self.temporary_directory,
            scene_walker_directory = os.path.join(self.location, 'paperCode'),
            execute = self.execute,
            variables = ['last_us'],
            params=self.params,
            verbose=verbose,
            model=self._model_function
        )

        conditional_log_densities = []

        #print("aggregating conditionial densities")
        for i in tqdm(range(len(fixations.x)), disable=not verbose):
            position = data_locations[i]
            conditional_density = results[position['image_no']][position['scanpath_no']]['last_us']  # [:, :, -2]
            conditional_log_density = np.log(conditional_density)
            conditional_log_density -= logsumexp(conditional_log_density)
            conditional_log_densities.append(conditional_log_density)

        return conditional_log_densities

    def conditional_log_density(self, stimulus, x_hist, y_hist, t_hist, out=None, durations=None, last_duration=0.3):
        stimuli = Stimuli([stimulus.stimulus_data])
        #fixations = Fixations(
        #    x=np.array([x_hist[-1]]),
        #    y=np.array([y_hist[-1]]),
        #    t=np.array([t_hist[-1]]),
        #    x_hist=np.array([x_hist[:-1]]),
        #    y_hist=np.array([y_hist[:-1]]),
        #    t_hist=np.array([t_hist[:-1]]),
        #    n=np.array([0], int),
        #    subjects=np.array([0], int),
        #)
        fixations = Fixations(
            x=np.zeros((1, )),
            y=np.zeros((1, )),
            t=np.zeros((1, ))+t_hist[-1]+last_duration,
            x_hist=np.array([x_hist]),
            y_hist=np.array([y_hist]),
            t_hist=np.array([t_hist]),
            n=np.array([0], int),
            subjects=np.array([0], int),
        )
        return self._conditional_log_densities(stimuli, fixations)[0]

    def log_likelihoods(self, stimuli, fixations, verbose=False):
        log_likelihoods = np.empty(len(fixations.x))
        inds = range(len(fixations.x))
        with tqdm(total=len(fixations.x), disable=not verbose) as t:
            for ind_chunk in chunked(inds, 1000):
                _f = fixations[ind_chunk]
                conditional_log_densities = self._conditional_log_densities(stimuli, _f)
                for k, i in enumerate(ind_chunk):
                    log_likelihoods[i] = conditional_log_densities[k][fixations.y_int[i], fixations.x_int[i]]
                t.update(len(ind_chunk))

        return log_likelihoods


class SceneWalkerEngbert(SceneWalker):
    """The SceneWalker model

    .. seealso::
        TODO
    """

    def __init__(self,
                 base_model, pixel_per_dva, temporary_directory,
                 omega=10**(-4.1722)*100,
                 rho=10**(-0.8343)*100,
                 sigmaAttention=6.1036,
                 sigmaInhib=2.1319,
                 gamma=.3,
                 lambda_=1,
                 inhibStrength=1,
                 zeta=0.04,
                 location=None,
                 execute=execute,
                 **kwargs):
        super(SceneWalkerEngbert, self).__init__(
            base_model,
            pixel_per_dva,
            temporary_directory,
            omega=omega,
            rho=rho,
            sigmaAttention=sigmaAttention,
            sigmaInhib=sigmaInhib,
            gamma=gamma,
            lambda_=lambda_,
            inhibStrength=inhibStrength,
            zeta=zeta,
            location=location,
            execute=execute,
            **kwargs
        )


class SceneWalkerDivisive(SceneWalker):
    """The SceneWalker model

    .. seealso::
        TODO
    """

    def __init__(self,
                 base_model, pixel_per_dva, temporary_directory,
                 omega=np.exp(0.1795),
                 rho=np.exp(6.7791),
                 sigmaAttention=np.exp(1.5225)*np.exp(0.4028),
                 sigmaInhib=np.exp(1.6466)*np.exp(2.3420),
                 gamma=np.exp(2.3420),
                 lambda_=np.exp(0.4028),
                 inhibStrength=np.exp(-9.8428),
                 zeta=np.exp(-3.0874),
                 location=None,
                 execute=execute,
                 **kwargs):
        super(SceneWalkerDivisive, self).__init__(
            base_model,
            pixel_per_dva,
            temporary_directory,
            omega=omega,
            rho=rho,
            sigmaAttention=sigmaAttention,
            sigmaInhib=sigmaInhib,
            gamma=gamma,
            lambda_=lambda_,
            inhibStrength=inhibStrength,
            zeta=zeta,
            location=location,
            execute=execute,
            **kwargs
        )
        self._model_function = 'dynamic_model_divisive'



from __future__ import absolute_import, print_function, division, unicode_literals

try:
    import matplotlib.pyplot as plt
except ImportError:
    # If matplotlib is not there, just ignore it
    pass

import numpy as np
from scipy.ndimage import zoom


def plot_information_gain(information_gain, ax=None, color_range = None, image=None, frame=False,
                          thickness = 1.0, zoom_factor=1.0, threshold=0.05, rel_levels=None,
                          alpha=0.5, color_offset = 0.25, plot_color_bar=True):
    """
    Create pixel space information gain plots as in the paper.

    Parameters:
    -----------

    information gain: the information gain to plot.
    ax: the matplotlib axes object to use. If none, use current axes.
    color_range: Full range of colorbar
    """
    if ax is None:
        ax = plt.gca()
    ig = information_gain

    if zoom_factor != 1.0:
        ig = zoom(ig, zoom_factor, order=0)

    if color_range is None:
        color_range = (ig.min(), ig.max())
    if not isinstance(color_range, (tuple, list)):
        color_range = (-color_range, color_range)

    color_total_max = max(np.abs(color_range[0]), np.abs(color_range[1]))

    if image is not None:
        if image.ndim == 3:
            image = image.sum(axis=-1)
        ax.imshow(image, alpha=0.3)

    if rel_levels is None:
        rel_levels = [0.1, 0.4, 0.7]

    # from https://stackoverflow.com/questions/8580631/transparent-colormap
    cm = plt.cm.get_cmap('RdBu')
    cm._init()
    alphas = (np.abs(np.linspace(-1.0, 1.0, cm.N)))
    alphas = np.ones_like(alphas)*alpha
    cm._lut[:-3, -1] = alphas

    levels = []
    colors = []

    min_val = np.abs(ig.min())
    max_val = np.abs(ig.max())

    total_max = max(min_val, max_val)

    def get_color(val):
        # value relative -1 .. 1
        rel_val = val / color_total_max
        # shift around 0
        rel_val = (rel_val + np.sign(rel_val) * color_offset) / (1+color_offset)
        # transform to 0 .. 1
        rel_val = (0.5 + rel_val / 2)
        return cm(rel_val)

    if min_val / total_max > threshold:
        for l in [1.0]+rel_levels[::-1]:
            val = -l*min_val
            levels.append(val)

            colors.append(get_color(val))
    else:
        levels.append(-total_max)
        colors.append('white')

    # We want to use the color from the value nearer to zero
    colors = colors[1:]
    colors.append((1.0, 1.0, 1.0, 0.0))

    if max_val / total_max > threshold:
        for l in rel_levels+[1.0]:
            val = l*max_val
            levels.append(val)

            colors.append(get_color(val))
    else:
        levels.append(total_max)

    #print rel_vals
    ax.contourf(ig, levels=levels,
                colors=colors,
                vmin=-color_total_max, vmax=color_total_max
                )
    ax.contour(ig, levels=levels,
               # colors=colors,
               #          vmin=-color_range, vmax=color_range
               colors = 'gray',
               linestyles='solid',
               linewidths=0.6*thickness
               )

    if plot_color_bar:
        ## Draw color range bar
        h = 100
        w = 10
        t = np.empty((100, 10, 4))
        for y in range(h):
            for x in range(w):
                val = (y/h) * (color_range[1] - color_range[0]) + color_range[0]
                color = np.asarray(get_color(val))
                if not -min_val <= val <= max_val:
                    color[-1] *= 0.4
                else:
                    color[-1] = 1
                t[y, x, :] = color

        ax.imshow(t, extent=(0.95*ig.shape[1], 0.98*ig.shape[1],
                             0.1*ig.shape[0], 0.9*ig.shape[0]))

    ax.set_xlim(0, ig.shape[1])
    ax.set_ylim(ig.shape[0], 0)

    if frame:
        # Just a frame
        ax.set_xticks([])
        ax.set_yticks([])
        [i.set_linewidth(i.get_linewidth()*thickness) for i in ax.spines.itervalues()]
    else:
        ax.set_axis_off()


def normalize_log_density(log_density):
    """ convertes a log density into a map of the cummulative distribution function.
    """
    density = np.exp(log_density)
    flat_density = density.flatten()
    inds = flat_density.argsort()[::-1]
    sorted_density = flat_density[inds]
    cummulative = np.cumsum(sorted_density)
    unsorted_cummulative = cummulative[np.argsort(inds)]
    return unsorted_cummulative.reshape(log_density.shape)

def visualize_distribution(log_densities, ax = None):
    if ax is None:
        ax = plt.gca()
    t = normalize_log_density(log_densities)
    img = ax.imshow(t, cmap=plt.cm.viridis)
    levels = levels=[0, 0.25, 0.5, 0.75, 1.0]
    cs = ax.contour(t, levels=levels, colors='black')
    #plt.clabel(cs)

    return img, cs



"""
Code to apply quilt patches to files

This module enables pysaliency to use quilt patches
to patch code from external saliency models. While
in Linux, quilt itself could be used to apply the patches,
in Windows and Mac quilt might not be available and
nontrivial to install for users.

It does not support all possible patch files but only
the subset of functionality needed by pysaliency.
"""

from __future__ import absolute_import, print_function, division, unicode_literals

import os.path

from .utils import full_split


class Hunk(object):
    def __init__(self, lines):
        meta_data = lines.pop(0)
        a, src_data, target_data, b = meta_data.split()
        assert a == '@@'
        assert b == '@@'
        start, length = self.parse_position(src_data)
        assert start < 0
        self.source_start = -start
        self.source_length = length

        start, length = self.parse_position(target_data)
        assert start > 0
        self.target_start = start
        self.target_length = length

        self.lines = lines

    def parse_position(self, position):
        start, length = position.split(',')
        start = int(start)
        length = int(length)
        return start, length

    def apply(self, source, target):
        src_pos = self.source_start - 1
        assert len(target) == self.target_start - 1
        for line in self.lines:
            type, data = line[0], line[1:]
            if type == ' ':
                assert source[src_pos] == data
                target.append(data)
                src_pos += 1
            elif type == '-':
                assert source[src_pos] == data
                src_pos += 1
            elif type == '+':
                target.append(data)
            elif type == '\\':
                # Newline stuff, ignore
                pass
            else:
                raise ValueError(line)
        assert src_pos == self.source_start + self.source_length - 1
        assert len(target) == self.target_start + self.target_length - 1
        return src_pos


class Diff(object):
    def __init__(self, lines):
        source = lines.pop(0)
        assert source.startswith('--- ')
        _, source = source.split('--- ', 1)
        source, _ = source.split('\t', 1)
        source = os.path.join(*full_split(source)[1:])
        target = lines.pop(0)
        assert target.startswith('+++ ')
        _, target = target.split('+++ ', 1)
        target, _ = target.split('\t', 1)
        target = os.path.join(*full_split(target)[1:])
        self.source_filename = source
        self.target_filename = target
        self.hunks = []
        while lines:
            assert lines[0].startswith('@@ ')
            hunk_lines = [lines.pop(0)]
            while lines and not lines[0].startswith('@@ '):
                line = lines.pop(0)
                if line:
                    hunk_lines.append(line)
            self.hunks.append(Hunk(hunk_lines))

    def apply(self, location):
        hunks = list(self.hunks)
        source = open(os.path.join(location, self.source_filename)).read()
        source = source.split('\n')
        target = []
        src_pos = 0
        while src_pos < len(source):
            if hunks:
                if hunks[0].source_start == src_pos+1:
                    hunk = hunks.pop(0)
                    src_pos = hunk.apply(source, target)
                    continue
            target.append(source[src_pos])
            src_pos += 1
        open(os.path.join(location, self.target_filename), 'w').write('\n'.join(target))


class PatchFile(object):
    def __init__(self, patch):
        self.diffs = []
        lines = patch.split('\n')
        while lines:
            index1 = lines.pop(0)
            assert index1.startswith('Index: ')
            index2 = lines.pop(0)
            assert index2.startswith('==============')
            diff = []
            diff.append(lines.pop(0))
            while lines and not lines[0].startswith('Index: '):
                diff.append(lines.pop(0))
            diff_obj = Diff(diff)
            self.diffs.append(diff_obj)

    def apply(self, location, verbose=True):
        for diff in self.diffs:
            if verbose:
                print("Patching {}".format(diff.source_filename))
            diff.apply(location)


class QuiltSeries(object):
    def __init__(self, patches_location):
        self.patches_location = patches_location
        series = open(os.path.join(self.patches_location, 'series')).read()
        self.patches = []
        self.patch_names = []
        for line in series.split('\n'):
            if not line:
                continue
            patch_content = open(os.path.join(self.patches_location, line)).read()
            self.patches.append(PatchFile(patch_content))
            self.patch_names.append(line)

    def apply(self, location, verbose=True):
        for patch, name in zip(self.patches, self.patch_names):
            if verbose:
                print("Applying {}".format(name))
            patch.apply(location, verbose=verbose)



from __future__ import print_function, absolute_import, division, unicode_literals

from abc import ABCMeta, abstractmethod
from six import add_metaclass

from .utils import remove_trailing_nans


@add_metaclass(ABCMeta)
class SamplingModelMixin(object):
    """A sampling model is supports sampling fixations and whole scanpaths."""
    def sample_scanpath(
        self, stimulus, x_hist, y_hist, t_hist, samples, attributes=None, verbose=False, rst=None
    ):
        """return xs, ys, ts"""
        xs = list(remove_trailing_nans(x_hist))
        ys = list(remove_trailing_nans(y_hist))
        ts = list(remove_trailing_nans(t_hist))
        if not len(xs) == len(ys) == len(ts):
            raise ValueError("Histories for x, y and t have to be the same length")

        for i in range(samples):
            x, y, t = self.sample_fixation(stimulus, xs, ys, ts, attributes=attributes, verbose=verbose, rst=rst)
            xs.append(x)
            ys.append(y)
            ts.append(t)

        return xs, ys, ts

    @abstractmethod
    def sample_fixation(self, stimulus, x_hist, y_hist, t_hist, attributes=None, verbose=False, rst=None):
        """return x, y, t"""
        raise NotImplementedError()


class ScanpathSamplingModelMixin(SamplingModelMixin):
    """A sampling model which only has to implement sample_scanpath instead of sample_fixation"""
    @abstractmethod
    def sample_scanpath(
            self, stimulus, x_hist, y_hist, t_hist, samples, attributes=None, verbose=False, rst=None
    ):
        raise NotImplementedError()

    def sample_fixation(self, stimulus, x_hist, y_hist, t_hist, attributes=None, verbose=False, rst=None):
        samples = 1
        xs, ys, ts = self.sample_scanpath(stimulus, x_hist, y_hist, t_hist, samples, attributes=attributes,
                                          verbose=verbose, rst=rst)
        return xs[-1], ys[-1], ts[-1]



from __future__ import print_function, unicode_literals, division, absolute_import

import numba


@numba.jit(nopython=True)
def fill_fixation_map(fixation_map, fixations):
    """fixationmap: 2d array. fixations: Nx2 array of y, x positions"""
    for i in range(len(fixations)):
        fixation_y, fixation_x = fixations[i]
        fixation_map[int(fixation_y), int(fixation_x)] += 1



from __future__ import absolute_import, print_function, division, unicode_literals

import os
import tempfile
import zipfile
import tarfile
from pkg_resources import resource_string,  resource_listdir

from boltons.fileutils import mkdir_p
import numpy as np
from scipy.ndimage import zoom

from .utils import TemporaryDirectory, download_and_check, run_matlab_cmd
from .saliency_map_models import SaliencyMapModel
from .external_models import ExternalModelMixin, download_extract_patch


class OpenSALICON(ExternalModelMixin, SaliencyMapModel):
    """OpenSALICON is an open source implementation of the SALICON saliency model

    .. seealso::
        Huang, X., Shen, C., Boix, X., & Zhao, Q. (2015). SALICON: Reducing
        the Semantic Gap in Saliency Prediction by Adapting Deep Neural
        Networks. In Proceedings of the IEEE International Conference on
        Computer Vision (pp. 262-270).

        https://github.com/CLT29/OpenSALICON
    """

    __modelname__ = 'OpenSALICON'

    def __init__(self, location=None, **kwargs):
        self.setup(location)
        super(OpenSALICON, self).__init__(**kwargs)
        import caffe
        prototxt_file = os.path.join(self.location, 'salicon.prototxt').encode('utf8')
        model_file = os.path.join(self.location, 'model_files', 'salicon_osie.caffemodel').encode('utf8')

        self.net = caffe.Net(prototxt_file, model_file, caffe.TEST)

    def _setup(self):
        mkdir_p(self.location)
        download_extract_patch('http://www.cs.pitt.edu/~chris/files/2016/model_files.tgz',
                               'c150041a975e52cf2ff234975a699fd6',
                               self.location,
                               location_in_archive=False,
                               patches=None)
        download_and_check('https://raw.githubusercontent.com/CLT29/OpenSALICON/36d4de26c8a57a4cfa27387df4f64a208a748e46/salicon.prototxt',
                           os.path.join(self.location, 'salicon.prototxt'),
                           'f7df84bd721b33d0c07518e466021037')

    def process_the_image(self, im):
        MEAN_VALUE = np.array([103.939, 116.779, 123.68])   # BGR
        MEAN_VALUE = MEAN_VALUE[:,None, None]
        # put channel dimension first
        im = np.transpose(im, (2,0,1))
        # switch to BGR
        im = im[::-1, :, :]
        # subtract mean
        im = im - MEAN_VALUE
        im = im[None,:]
        im = im / 255 # convert to float precision
        return im

    def _saliency_map(self, stimulus):
        FINE_SCALE = np.array([1,3,1200,1600], dtype=np.float32)
        COARSE_SCALE = np.array([1,3,600,800], dtype=np.float32)
        im = self.process_the_image(stimulus)
        coarse_img = zoom(im,tuple(COARSE_SCALE / np.asarray(im.shape, dtype=np.float32)), np.dtype(np.float32), mode='nearest')
        assert(coarse_img.shape == (1,3,600,800))
        fine_img = zoom(im,tuple(FINE_SCALE / np.asarray(im.shape, dtype=np.float32)), np.dtype(np.float32), mode='nearest')
        assert(fine_img.shape == (1,3,1200,1600))
        self.net.blobs['fine_scale'].data[...] = fine_img
        self.net.blobs['coarse_scale'].data[...] = coarse_img
        self.net.forward()
        sal_map = self.net.blobs['saliency_map_out'].data
        sal_map = sal_map[0,0,:,:]
        sal_map = sal_map - np.amin(sal_map)
        sal_map = sal_map / np.amax(sal_map)
        sal_map = zoom(sal_map,tuple(np.asarray(im.shape[2:], dtype=np.float32) / np.asarray(sal_map.shape,dtype=np.float32)), np.dtype(np.float32), mode='nearest')
        return sal_map





from __future__ import print_function, absolute_import, division
from collections import Sequence, MutableMapping
from itertools import chain
from glob import iglob

import warnings as _warnings
import os as _os
import sys as _sys
import os
import hashlib
from functools import partial
import warnings
from six.moves import urllib, filterfalse, map
from six import iterkeys
import subprocess as sp
from tempfile import mkdtemp

import numpy as np

from boltons.cacheutils import LRU
import deprecation
from tqdm import tqdm
import requests


def full_split(filename):
    """ split filename into all parts """
    path, basename = os.path.split(filename)
    components = [path, basename]
    while components[0] and components[0] != '/':
        first_part = components.pop(0)
        path, basename = os.path.split(first_part)
        components = [path, basename] + components

    if components[0] == '':
        components = components[1:]

    return components


def get_minimal_unique_filenames(filenames):
    if len(filenames) <= 1:
        return [os.path.basename(item) for item in filenames]

    components = [full_split(filename) for filename in filenames]

    while len(set(item[0] for item in components)) <= 1:
        components = [item[1:] for item in components]

    return [os.path.join(*item) for item in components]


def remove_trailing_nans(data):
    """Filters a scanpath arrays to remove the ending part of nans."""
    for i in range(len(data)):
        if np.all(np.isnan(data[i:])):
            return data[:i]
    return data


def lazy_property(fn):
    """Lazy property: Is only calculated when first used.
       Code from http://stackoverflow.com/questions/3012421/python-lazy-property-decorator"""
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazyprop(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazyprop


class LazyList(Sequence):
    """
    A list-like class that is able to generate it's entries only
    when needed. Entries can be cached.

    LayList implements `collections.Sequence` and therefore `__contains__`.
    However, to use it, in the worst case all elements of the list have
    to be generated.

    .. note::
        As `LazyList` stores the generator function, pickling it
        will usually fail. To pickle a `LazyList`, use `dill`.
    """
    def __init__(self, generator, length, cache=True, pickle_cache=False):
        """
        Parameters
        ----------

        @type  generator: callable
        @param generator: A function that takes an integer `n` and returns the
                          `n`-th element of the list.

        @type  length:   int
        @param length:   The length of the list

        @type  cache: bool, defaults to `True`
        @param cache: Wether to cache the list items.

        @type  pickle_cache: bool, defaults to `False`
        @param pickle_cache: Whether the cache should be saved when
                             pickling the object.
        """
        self.generator = generator
        self.length = length
        self.cache = cache
        self.pickle_cache = pickle_cache
        self._cache = {}

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [self[i] for i in range(len(self))[index]]
        elif isinstance(index, list):
            return [self[i] for i in index]
        else:
            return self._getitem(index)

    def _getitem(self, index):
        if not 0 <= index < self.length:
            raise IndexError(index)
        if index in self._cache:
            return self._cache[index]
        value = self.generator(index)
        if self.cache:
            self._cache[index] = value
        return value

    def __getstate__(self):
        # we don't want to save the cache
        state = dict(self.__dict__)
        if not self.pickle_cache:
            state.pop('_cache')
        return state

    def __setstate__(self, state):
        if not '_cache' in state:
            state['_cache'] = {}
        self.__dict__ = dict(state)


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.

    From http://stackoverflow.com/a/19299884
    """

    def __init__(self, suffix="", prefix="tmp", dir=None, cleanup=True):
        self._closed = False
        self.name = None  # Handle mkdtemp raising an exception
        self.name = mkdtemp(suffix, prefix, dir)
        self.do_cleanup = cleanup

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if not self.do_cleanup:
            return
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                print("ERROR: {!r} while cleaning up {!r}".format(ex, self,),
                      file=_sys.stderr)
                return
            self._closed = True
            if _warn:
                self._warn("Implicitly cleaning up {!r}".format(self))

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass


def which(program):
    """
    Check whether a program is present on the system.
    from https://stackoverflow.com/a/377028
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def filter_files(filenames, ignores):
    """
    Filter a list of files, excluding all filenames which contain
    an element of `ignores` as part of their path
    """
    parts = map(full_split, filenames)
    inds = [i for i, ps in enumerate(parts)
            if not any([ignore in ps for ignore in ignores])]
    return [filenames[i] for i in inds]


class MatlabOptions(object):
    matlab_names = ['matlab', 'matlab.exe']
    octave_names = ['octave', 'octave.exe']


def get_matlab_or_octave():
    for name in MatlabOptions.matlab_names + MatlabOptions.octave_names:
        if which(name):
            return which(name)
    raise Exception('No version of matlab or octave was found on this system!')


def run_matlab_cmd(cmd, cwd=None):
    matlab = get_matlab_or_octave()
    args = []
    if os.path.basename(matlab).startswith('matlab'):
        args += ['-nodesktop', '-nosplash', '-r']
        args.append("try;{};catch exc;disp(getReport(exc));disp('__ERROR__');exit(1);end;quit".format(cmd))
    else:
        args += ['--traditional', '--eval']
        args.append("try;{};catch exc;struct_levels_to_print(10);print_struct_array_contents(true);disp(lasterror);for i=1:size(lasterror.stack);disp(lasterror.stack(i));end;disp('__ERROR__');exit(1);end;quit".format(cmd))
    sp.check_call([matlab] + args, cwd=cwd)


def check_file_hash(filename, md5_hash):
    """
    Check a file's hash and issue a warning it is has not the expected value.
    """
    print('Checking md5 sum...')
    hasher = hashlib.md5()
    with open(filename, 'rb') as f:
        # read file in chunks to avoid "invalid argument" error
        # see https://stackoverflow.com/a/48123430
        for block in iter(partial(f.read, 64 * (1 << 20)), b''):
            hasher.update(block)

    file_hash = hasher.hexdigest()
    if file_hash != md5_hash:
        warnings.warn("MD5 sum of {} has changed. Expected {} but got {}. This might lead to"
                      " this code producing wrong data.".format(filename, md5_hash, file_hash))


def download_file_old(url, target):
    """Download url to target while displaying progress information."""
    class Log(object):
        def __init__(self):
            self.last_percent = -1

        def __call__(self, blocks_recieved, block_size, file_size):
            percent = int(blocks_recieved * block_size / file_size * 100)
            if percent == self.last_percent:
                return
            print('\rDownloading file. {}% done'.format(percent), end='')
            self.last_percent = percent
    urllib.request.urlretrieve(url, target, Log())
    print('')


def download_file(url, target):
    r = requests.get(url, stream=True)
    total_size = int(r.headers.get('content-length', 0))
    with open(target, 'wb') as f:
        with tqdm(total=total_size, unit='B', unit_scale=True, desc='Downloading file') as progress_bar:
            for data in r.iter_content(32*1024):
                f.write(data)
                progress_bar.update(32*1024)


def download_and_check(url, target, md5_hash):
    """Download url to target and check for correct md5_hash. Prints warning if hash is not correct."""
    download_file(url, target)
    check_file_hash(target, md5_hash)


def download_file_from_google_drive(id, destination):
    """adapted from https://drive.google.com/uc?id=0B2hsWbciDVedWHFiMUVVWFRZTE0&export=download"""
    def get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value

        return None

    def save_response_content(response, destination):
        CHUNK_SIZE = 32768

        with tqdm(unit='B', unit_scale=True) as pbar:
            with open(destination, "wb") as f:
                for chunk in response.iter_content(CHUNK_SIZE):
                    if chunk: # filter out keep-alive new chunks
                        f.write(chunk)
                        pbar.update(CHUNK_SIZE)

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)


class Cache(MutableMapping):
    """Cache that supports saving the items to files

    Set `cache_location` to save all newly set
    items to .npy files in cache_location.

    .. warning ::
        Items that have been set before setting `cache_location` won't
        be saved to files!

    """
    def __init__(self, cache_location=None, pickle_cache=False,
                 memory_cache_size=None):
        if memory_cache_size:
            self._cache = LRU(max_size=memory_cache_size)
        else:
            self._cache = {}
        self.cache_location = cache_location
        self.pickle_cache = pickle_cache

    def clear(self):
        """ Clear memory cache"""
        self._cache = {}

    def filename(self, key):
        return os.path.join(self.cache_location, '{}.npy'.format(key))

    def __getitem__(self, key):
        if not key in self._cache:
            if self.cache_location is not None:
                filename = self.filename(key)
                if os.path.exists(filename):
                    value = np.load(filename)
                    self._cache[key] = value
                else:
                    raise KeyError('Key {} neither in cache nor on disk'.format(key))
        return self._cache[key]

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError('Only string keys are supported right now!')
        if self.cache_location is not None:
            if not os.path.exists(self.cache_location):
                os.makedirs(self.cache_location)
            filename = self.filename(key)
            np.save(filename, value)
        self._cache[key] = value

    def __delitem__(self, key):
        if self.cache_location is not None:
            filename = self.filename(key)
            if os.path.exists(filename):
                os.remove(filename)
        del self._cache[key]

    def __iter__(self):
        if self.cache_location is not None:
            filenames = iglob(self.filename('*'))
            keys = map(lambda f: os.path.splitext(os.path.basename(f))[0], filenames)
            new_keys = filterfalse(lambda key: key in self._cache.keys(), keys)
            return chain(iterkeys(self._cache), new_keys)
        else:
            return iterkeys(self._cache)

    def __len__(self):
        i = iter(self)
        return len(list(i))

    def __getstate__(self):
        # we don't want to save the cache
        state = dict(self.__dict__)
        if not self.pickle_cache:
            state.pop('_cache')
        return state

    def __setstate__(self, state):
        if not '_cache' in state:
            if state.get('memory_cache_size'):
                state['_cache'] = LRU(max_size=memory_cache_size)
            else:
                state['_cache'] = {}
        self.__dict__ = dict(state)


def average_values(values, fixations, average='fixation'):
    if average == 'fixation':
        return np.mean(values)
    elif average == 'image':
        import pandas as pd
        df = pd.DataFrame({'n': fixations.n, 'value': values})
        return df.groupby('n')['value'].mean().mean()
    else:
        raise ValueError(average)


def deprecated_class(deprecated_in=None, removed_in=None, current_version=None, details=''):
    def wrap(cls):
        class DeprecatedClass(cls):
            @deprecation.deprecated(deprecated_in=deprecated_in, removed_in=removed_in, current_version=current_version, details=details)
            def __init__(self, *args, **kwargs):
                super(DeprecatedClass, self).__init__(*args, **kwargs)

        return DeprecatedClass

    return wrap



from __future__ import absolute_import, print_function, division, unicode_literals

import numpy as np


def normalize_saliency_map(saliency_map, cdf, cdf_bins):
    """ Normalize saliency to make saliency values distributed according to a given CDF
    """

    smap = saliency_map.copy()
    shape = smap.shape
    smap = smap.flatten()
    smap = np.argsort(np.argsort(smap)).astype(float)
    smap /= 1.0*len(smap)

    inds = np.searchsorted(cdf, smap, side='right')
    smap = cdf_bins[inds]
    smap = smap.reshape(shape)
    smap = smap.reshape(shape)
    return smap


def NSS(saliency_map, xs, ys):
    xs = np.asarray(xs, dtype=np.int)
    ys = np.asarray(ys, dtype=np.int)

    mean = saliency_map.mean()
    std = saliency_map.std()

    value = saliency_map[ys, xs].copy()
    value -= mean

    if std:
        value /= std

    return value
